"""Tests for style operation handlers — style, border, define-style, apply-style."""

from __future__ import annotations

import pytest
from copy import copy
from fcp_core import ParsedOp

from fcp_sheets.adapter import SheetsAdapter
from fcp_sheets.model.snapshot import SheetsModel
from fcp_sheets.server.ops_style import op_style, op_border, op_define_style, op_apply_style
from fcp_sheets.server.resolvers import SheetsOpContext


# ── Style verb ───────────────────────────────────────────────────────────────


class TestStyleBold:
    def test_bold_single_cell(self, ctx: SheetsOpContext):
        op = ParsedOp(verb="style", positionals=["A1", "bold"], raw="style A1 bold")
        result = op_style(op, ctx)
        assert result.success
        cell = ctx.wb.active.cell(row=1, column=1)
        assert cell.font.bold is True

    def test_bold_range(self, ctx: SheetsOpContext):
        op = ParsedOp(verb="style", positionals=["A1:B2", "bold"], raw="style A1:B2 bold")
        result = op_style(op, ctx)
        assert result.success
        for r in (1, 2):
            for c in (1, 2):
                assert ctx.wb.active.cell(row=r, column=c).font.bold is True


class TestStyleItalic:
    def test_italic(self, ctx: SheetsOpContext):
        op = ParsedOp(verb="style", positionals=["C3", "italic"], raw="style C3 italic")
        result = op_style(op, ctx)
        assert result.success
        assert ctx.wb.active.cell(row=3, column=3).font.italic is True


class TestStyleUnderline:
    def test_underline(self, ctx: SheetsOpContext):
        op = ParsedOp(verb="style", positionals=["A1", "underline"], raw="style A1 underline")
        result = op_style(op, ctx)
        assert result.success
        assert ctx.wb.active.cell(row=1, column=1).font.underline == "single"


class TestStyleStrike:
    def test_strikethrough(self, ctx: SheetsOpContext):
        op = ParsedOp(verb="style", positionals=["A1", "strike"], raw="style A1 strike")
        result = op_style(op, ctx)
        assert result.success
        assert ctx.wb.active.cell(row=1, column=1).font.strike is True


class TestStyleFont:
    def test_font_name(self, ctx: SheetsOpContext):
        op = ParsedOp(
            verb="style", positionals=["A1"],
            params={"font": "Helvetica"}, raw="style A1 font:Helvetica",
        )
        result = op_style(op, ctx)
        assert result.success
        assert ctx.wb.active.cell(row=1, column=1).font.name == "Helvetica"

    def test_font_size(self, ctx: SheetsOpContext):
        op = ParsedOp(
            verb="style", positionals=["A1"],
            params={"size": "14"}, raw="style A1 size:14",
        )
        result = op_style(op, ctx)
        assert result.success
        assert ctx.wb.active.cell(row=1, column=1).font.size == 14.0


class TestStyleColor:
    def test_hex_color(self, ctx: SheetsOpContext):
        op = ParsedOp(
            verb="style", positionals=["A1"],
            params={"color": "#FF0000"}, raw="style A1 color:#FF0000",
        )
        result = op_style(op, ctx)
        assert result.success
        cell = ctx.wb.active.cell(row=1, column=1)
        # openpyxl stores color as a Color object; font.color can be checked
        assert cell.font.color is not None

    def test_named_color(self, ctx: SheetsOpContext):
        op = ParsedOp(
            verb="style", positionals=["A1"],
            params={"color": "blue"}, raw="style A1 color:blue",
        )
        result = op_style(op, ctx)
        assert result.success
        assert ctx.wb.active.cell(row=1, column=1).font.color is not None


class TestStyleFill:
    def test_fill_color(self, ctx: SheetsOpContext):
        op = ParsedOp(
            verb="style", positionals=["A1"],
            params={"fill": "#FFFF00"}, raw="style A1 fill:#FFFF00",
        )
        result = op_style(op, ctx)
        assert result.success
        cell = ctx.wb.active.cell(row=1, column=1)
        assert cell.fill.start_color.rgb is not None
        assert cell.fill.fill_type == "solid"

    def test_fill_named_color(self, ctx: SheetsOpContext):
        op = ParsedOp(
            verb="style", positionals=["A1"],
            params={"fill": "green"}, raw="style A1 fill:green",
        )
        result = op_style(op, ctx)
        assert result.success
        cell = ctx.wb.active.cell(row=1, column=1)
        assert cell.fill.fill_type == "solid"


class TestStyleAlignment:
    def test_align_center(self, ctx: SheetsOpContext):
        op = ParsedOp(
            verb="style", positionals=["A1"],
            params={"align": "center"}, raw="style A1 align:center",
        )
        result = op_style(op, ctx)
        assert result.success
        assert ctx.wb.active.cell(row=1, column=1).alignment.horizontal == "center"

    def test_align_right(self, ctx: SheetsOpContext):
        op = ParsedOp(
            verb="style", positionals=["A1"],
            params={"align": "right"}, raw="style A1 align:right",
        )
        result = op_style(op, ctx)
        assert result.success
        assert ctx.wb.active.cell(row=1, column=1).alignment.horizontal == "right"

    def test_align_left(self, ctx: SheetsOpContext):
        op = ParsedOp(
            verb="style", positionals=["A1"],
            params={"align": "left"}, raw="style A1 align:left",
        )
        result = op_style(op, ctx)
        assert result.success
        assert ctx.wb.active.cell(row=1, column=1).alignment.horizontal == "left"

    def test_valign_top(self, ctx: SheetsOpContext):
        op = ParsedOp(
            verb="style", positionals=["A1"],
            params={"valign": "top"}, raw="style A1 valign:top",
        )
        result = op_style(op, ctx)
        assert result.success
        assert ctx.wb.active.cell(row=1, column=1).alignment.vertical == "top"

    def test_valign_middle(self, ctx: SheetsOpContext):
        op = ParsedOp(
            verb="style", positionals=["A1"],
            params={"valign": "middle"}, raw="style A1 valign:middle",
        )
        result = op_style(op, ctx)
        assert result.success
        # "middle" is mapped to "center" for openpyxl compatibility
        assert ctx.wb.active.cell(row=1, column=1).alignment.vertical == "center"


class TestStyleWrap:
    def test_wrap_text(self, ctx: SheetsOpContext):
        op = ParsedOp(verb="style", positionals=["A1", "wrap"], raw="style A1 wrap")
        result = op_style(op, ctx)
        assert result.success
        assert ctx.wb.active.cell(row=1, column=1).alignment.wrap_text is True


class TestStyleIndent:
    def test_indent(self, ctx: SheetsOpContext):
        op = ParsedOp(
            verb="style", positionals=["A1"],
            params={"indent": "2"}, raw="style A1 indent:2",
        )
        result = op_style(op, ctx)
        assert result.success
        assert ctx.wb.active.cell(row=1, column=1).alignment.indent == 2


class TestStyleRotate:
    def test_rotate(self, ctx: SheetsOpContext):
        op = ParsedOp(
            verb="style", positionals=["A1"],
            params={"rotate": "45"}, raw="style A1 rotate:45",
        )
        result = op_style(op, ctx)
        assert result.success
        assert ctx.wb.active.cell(row=1, column=1).alignment.text_rotation == 45


class TestStyleFormat:
    def test_number_format_raw(self, ctx: SheetsOpContext):
        op = ParsedOp(
            verb="style", positionals=["A1"],
            params={"fmt": "0.00%"}, raw="style A1 fmt:0.00%",
        )
        result = op_style(op, ctx)
        assert result.success
        assert ctx.wb.active.cell(row=1, column=1).number_format == "0.00%"

    def test_number_format_alias(self, ctx: SheetsOpContext):
        op = ParsedOp(
            verb="style", positionals=["A1"],
            params={"fmt": "currency"}, raw="style A1 fmt:currency",
        )
        result = op_style(op, ctx)
        assert result.success
        assert ctx.wb.active.cell(row=1, column=1).number_format == "$#,##0"


class TestStyleCombined:
    def test_bold_italic_with_font(self, ctx: SheetsOpContext):
        op = ParsedOp(
            verb="style", positionals=["A1", "bold", "italic"],
            params={"font": "Arial", "size": "12"},
            raw="style A1 bold italic font:Arial size:12",
        )
        result = op_style(op, ctx)
        assert result.success
        cell = ctx.wb.active.cell(row=1, column=1)
        assert cell.font.bold is True
        assert cell.font.italic is True
        assert cell.font.name == "Arial"
        assert cell.font.size == 12.0

    def test_full_styling(self, ctx: SheetsOpContext):
        op = ParsedOp(
            verb="style", positionals=["A1:B2", "bold", "wrap"],
            params={
                "font": "Calibri", "size": "11", "color": "#333333",
                "fill": "#EEEEEE", "align": "center", "valign": "middle",
                "fmt": "number2",
            },
            raw="style A1:B2 bold wrap font:Calibri size:11 color:#333333 fill:#EEEEEE align:center valign:middle fmt:number2",
        )
        result = op_style(op, ctx)
        assert result.success
        cell = ctx.wb.active.cell(row=1, column=1)
        assert cell.font.bold is True
        assert cell.font.name == "Calibri"
        assert cell.alignment.horizontal == "center"
        assert cell.alignment.wrap_text is True
        assert cell.number_format == "0.00"

    def test_preserves_existing_font_attrs(self, ctx: SheetsOpContext):
        """When applying bold, existing font name should be preserved."""
        from openpyxl.styles import Font
        cell = ctx.wb.active.cell(row=1, column=1)
        cell.font = Font(name="Courier New", size=16)

        op = ParsedOp(verb="style", positionals=["A1", "bold"], raw="style A1 bold")
        op_style(op, ctx)

        assert cell.font.bold is True
        assert cell.font.name == "Courier New"
        assert cell.font.size == 16.0


class TestStyleErrors:
    def test_no_args(self, ctx: SheetsOpContext):
        op = ParsedOp(verb="style", positionals=[], raw="style")
        result = op_style(op, ctx)
        assert not result.success

    def test_no_cells_resolved(self, ctx: SheetsOpContext):
        """Style with selector that matches no cells."""
        op = ParsedOp(
            verb="style", positionals=["bold"],
            selectors=["@type:formula"],
            raw="style bold @type:formula",
        )
        result = op_style(op, ctx)
        assert not result.success


class TestStyleWithSelectors:
    def test_style_with_range_selector(self, ctx: SheetsOpContext):
        """Style cells using @range selector."""
        ws = ctx.wb.active
        ws.cell(row=1, column=1, value="A")
        ws.cell(row=1, column=2, value="B")
        ctx.index.expand_bounds("Sheet1", 1, 1)
        ctx.index.expand_bounds("Sheet1", 1, 2)

        op = ParsedOp(
            verb="style", positionals=["bold"],
            selectors=["@range:A1:B1"],
            raw="style bold @range:A1:B1",
        )
        result = op_style(op, ctx)
        assert result.success
        assert ws.cell(row=1, column=1).font.bold is True
        assert ws.cell(row=1, column=2).font.bold is True


# ── Border verb ──────────────────────────────────────────────────────────────


class TestBorderAll:
    def test_border_all(self, ctx: SheetsOpContext):
        op = ParsedOp(
            verb="border", positionals=["A1:B2", "all"],
            raw="border A1:B2 all",
        )
        result = op_border(op, ctx)
        assert result.success
        cell = ctx.wb.active.cell(row=1, column=1)
        assert cell.border.top.style == "thin"
        assert cell.border.bottom.style == "thin"
        assert cell.border.left.style == "thin"
        assert cell.border.right.style == "thin"

    def test_border_all_cells_in_range(self, ctx: SheetsOpContext):
        op = ParsedOp(
            verb="border", positionals=["A1:C3", "all"],
            raw="border A1:C3 all",
        )
        result = op_border(op, ctx)
        assert result.success
        for r in range(1, 4):
            for c in range(1, 4):
                cell = ctx.wb.active.cell(row=r, column=c)
                assert cell.border.top.style is not None


class TestBorderOutline:
    def test_border_outline(self, ctx: SheetsOpContext):
        op = ParsedOp(
            verb="border", positionals=["A1:C3", "outline"],
            raw="border A1:C3 outline",
        )
        result = op_border(op, ctx)
        assert result.success
        ws = ctx.wb.active
        # Top-left corner: top and left borders
        assert ws.cell(row=1, column=1).border.top.style == "thin"
        assert ws.cell(row=1, column=1).border.left.style == "thin"
        # Bottom-right corner: bottom and right borders
        assert ws.cell(row=3, column=3).border.bottom.style == "thin"
        assert ws.cell(row=3, column=3).border.right.style == "thin"
        # Interior cell: no outline borders
        assert ws.cell(row=2, column=2).border.top.style is None
        assert ws.cell(row=2, column=2).border.left.style is None


class TestBorderSides:
    def test_border_top(self, ctx: SheetsOpContext):
        op = ParsedOp(
            verb="border", positionals=["A1", "top"],
            raw="border A1 top",
        )
        result = op_border(op, ctx)
        assert result.success
        assert ctx.wb.active.cell(row=1, column=1).border.top.style == "thin"

    def test_border_bottom(self, ctx: SheetsOpContext):
        op = ParsedOp(
            verb="border", positionals=["A1", "bottom"],
            raw="border A1 bottom",
        )
        result = op_border(op, ctx)
        assert result.success
        assert ctx.wb.active.cell(row=1, column=1).border.bottom.style == "thin"

    def test_border_left(self, ctx: SheetsOpContext):
        op = ParsedOp(
            verb="border", positionals=["A1", "left"],
            raw="border A1 left",
        )
        result = op_border(op, ctx)
        assert result.success
        assert ctx.wb.active.cell(row=1, column=1).border.left.style == "thin"

    def test_border_right(self, ctx: SheetsOpContext):
        op = ParsedOp(
            verb="border", positionals=["A1", "right"],
            raw="border A1 right",
        )
        result = op_border(op, ctx)
        assert result.success
        assert ctx.wb.active.cell(row=1, column=1).border.right.style == "thin"


class TestBorderLineStyle:
    def test_medium_border(self, ctx: SheetsOpContext):
        op = ParsedOp(
            verb="border", positionals=["A1", "all"],
            params={"line": "medium"}, raw="border A1 all line:medium",
        )
        result = op_border(op, ctx)
        assert result.success
        assert ctx.wb.active.cell(row=1, column=1).border.top.style == "medium"

    def test_thick_border(self, ctx: SheetsOpContext):
        op = ParsedOp(
            verb="border", positionals=["A1", "all"],
            params={"line": "thick"}, raw="border A1 all line:thick",
        )
        result = op_border(op, ctx)
        assert result.success
        assert ctx.wb.active.cell(row=1, column=1).border.top.style == "thick"

    def test_dashed_border(self, ctx: SheetsOpContext):
        op = ParsedOp(
            verb="border", positionals=["A1", "all"],
            params={"line": "dashed"}, raw="border A1 all line:dashed",
        )
        result = op_border(op, ctx)
        assert result.success
        assert ctx.wb.active.cell(row=1, column=1).border.top.style == "dashed"

    def test_invalid_line_style(self, ctx: SheetsOpContext):
        op = ParsedOp(
            verb="border", positionals=["A1", "all"],
            params={"line": "zigzag"}, raw="border A1 all line:zigzag",
        )
        result = op_border(op, ctx)
        assert not result.success


class TestBorderColor:
    def test_border_with_color(self, ctx: SheetsOpContext):
        op = ParsedOp(
            verb="border", positionals=["A1", "all"],
            params={"color": "#FF0000"}, raw="border A1 all color:#FF0000",
        )
        result = op_border(op, ctx)
        assert result.success
        cell = ctx.wb.active.cell(row=1, column=1)
        assert cell.border.top.color is not None


class TestBorderErrors:
    def test_no_args(self, ctx: SheetsOpContext):
        op = ParsedOp(verb="border", positionals=[], raw="border")
        result = op_border(op, ctx)
        assert not result.success

    def test_missing_sides(self, ctx: SheetsOpContext):
        op = ParsedOp(verb="border", positionals=["A1"], raw="border A1")
        result = op_border(op, ctx)
        assert not result.success


# ── Define-style / apply-style ───────────────────────────────────────────────


class TestDefineStyle:
    def test_define_basic_style(self, ctx: SheetsOpContext):
        op = ParsedOp(
            verb="define-style", positionals=["header", "bold"],
            params={"font": "Arial", "size": "14", "fill": "#4472C4", "color": "#FFFFFF"},
            raw="define-style header bold font:Arial size:14 fill:#4472C4 color:#FFFFFF",
        )
        result = op_define_style(op, ctx)
        assert result.success
        assert "header" in ctx.named_styles
        assert ctx.named_styles["header"]["bold"] is True
        assert ctx.named_styles["header"]["font"] == "Arial"

    def test_define_style_no_name(self, ctx: SheetsOpContext):
        op = ParsedOp(verb="define-style", positionals=[], raw="define-style")
        result = op_define_style(op, ctx)
        assert not result.success

    def test_define_style_with_flags(self, ctx: SheetsOpContext):
        op = ParsedOp(
            verb="define-style", positionals=["emphasis", "bold", "italic"],
            raw="define-style emphasis bold italic",
        )
        result = op_define_style(op, ctx)
        assert result.success
        assert ctx.named_styles["emphasis"]["bold"] is True
        assert ctx.named_styles["emphasis"]["italic"] is True


class TestApplyStyle:
    def test_apply_defined_style(self, ctx: SheetsOpContext):
        # First define a style
        define_op = ParsedOp(
            verb="define-style", positionals=["header", "bold"],
            params={"fill": "#4472C4"},
            raw="define-style header bold fill:#4472C4",
        )
        op_define_style(define_op, ctx)

        # Apply it
        apply_op = ParsedOp(
            verb="apply-style", positionals=["header", "A1:C1"],
            raw="apply-style header A1:C1",
        )
        result = op_apply_style(apply_op, ctx)
        assert result.success
        cell = ctx.wb.active.cell(row=1, column=1)
        assert cell.font.bold is True
        assert cell.fill.fill_type == "solid"

    def test_apply_unknown_style(self, ctx: SheetsOpContext):
        op = ParsedOp(
            verb="apply-style", positionals=["nonexistent", "A1"],
            raw="apply-style nonexistent A1",
        )
        result = op_apply_style(op, ctx)
        assert not result.success

    def test_apply_style_no_args(self, ctx: SheetsOpContext):
        op = ParsedOp(verb="apply-style", positionals=[], raw="apply-style")
        result = op_apply_style(op, ctx)
        assert not result.success

    def test_apply_style_with_fmt(self, ctx: SheetsOpContext):
        """Define and apply a style with number format."""
        define_op = ParsedOp(
            verb="define-style", positionals=["money"],
            params={"fmt": "currency2", "bold": "true"},
            raw="define-style money fmt:currency2 bold:true",
        )
        op_define_style(define_op, ctx)

        apply_op = ParsedOp(
            verb="apply-style", positionals=["money", "B2"],
            raw="apply-style money B2",
        )
        result = op_apply_style(apply_op, ctx)
        assert result.success



"""Chart operation handlers — chart add/series/axis/remove."""

from __future__ import annotations

from openpyxl.chart import Reference

from fcp_core import OpResult, ParsedOp

from fcp_sheets.lib.chart_types import get_chart_class
from fcp_sheets.model.refs import parse_cell_ref, parse_range_ref, RangeRef
from fcp_sheets.server.resolvers import SheetsOpContext


# Legend position aliases — LLMs often use full words
_LEGEND_POSITION_MAP: dict[str, str] = {
    "r": "r",
    "l": "l",
    "t": "t",
    "b": "b",
    "tr": "tr",
    "right": "r",
    "left": "l",
    "top": "t",
    "bottom": "b",
    "top-right": "tr",
    "topright": "tr",
}


def _get_title_text(title_obj) -> str | None:
    """Extract the plain text string from an openpyxl Title object.

    openpyxl wraps title strings in Title objects with nested rich text.
    This extracts the actual string from charts, axes, etc.
    """
    if title_obj is None:
        return None
    if isinstance(title_obj, str):
        return title_obj
    # Title object — extract text from rich text runs
    try:
        for p in title_obj.tx.rich.p:
            for r in p.r:
                return r.t
    except (AttributeError, TypeError):
        pass
    return None


def _get_chart_title_text(chart) -> str | None:
    """Extract the plain text title string from an openpyxl chart."""
    return _get_title_text(chart.title)


def _find_chart(ws, title: str):
    """Find a chart by title in the worksheet."""
    for chart in ws._charts:
        if _get_chart_title_text(chart) == title:
            return chart
    return None


def _parse_range_to_reference(ws, range_str: str) -> Reference | None:
    """Parse a range string like 'A1:D5' into an openpyxl Reference."""
    ref = parse_range_ref(range_str)
    if isinstance(ref, RangeRef):
        return Reference(
            worksheet=ws,
            min_col=ref.start.col,
            min_row=ref.start.row,
            max_col=ref.end.col,
            max_row=ref.end.row,
        )
    # Try single cell as a 1x1 range
    cell = parse_cell_ref(range_str)
    if cell:
        return Reference(
            worksheet=ws,
            min_col=cell.col,
            min_row=cell.row,
            max_col=cell.col,
            max_row=cell.row,
        )
    return None


def _chart_add(op: ParsedOp, ctx: SheetsOpContext) -> OpResult:
    """chart add TYPE [title:"TEXT"] data:RANGE [categories:RANGE] [at:CELL] [size:WxH] [legend:POS] [style:N]"""
    if len(op.positionals) < 2:
        return OpResult(
            success=False,
            message='Usage: chart add TYPE [title:"TEXT"] data:RANGE [categories:RANGE] [at:CELL] [size:WxH]',
        )

    chart_type = op.positionals[1]

    # Look up chart class
    try:
        cls, grouping, type_override = get_chart_class(chart_type)
    except ValueError as e:
        return OpResult(success=False, message=str(e))

    # Data range is required
    data_range = op.params.get("data")
    if not data_range:
        return OpResult(success=False, message="Missing required param: data:RANGE")

    ws = ctx.active_sheet

    # Create chart instance
    chart = cls()

    # Set title
    title = op.params.get("title")
    if title:
        chart.title = title

    # Set grouping/type override
    if grouping is not None:
        chart.grouping = grouping
    if type_override is not None:
        chart.type = type_override

    # Set style
    style_num = op.params.get("style")
    if style_num:
        try:
            chart.style = int(style_num)
        except (ValueError, TypeError):
            pass

    # Parse data range
    data_ref = _parse_range_to_reference(ws, data_range)
    if data_ref is None:
        return OpResult(success=False, message=f"Invalid data range: {data_range!r}")

    chart.add_data(data_ref, titles_from_data=True)

    # Parse categories range
    cat_range = op.params.get("categories")
    if cat_range:
        cat_ref = _parse_range_to_reference(ws, cat_range)
        if cat_ref is None:
            return OpResult(success=False, message=f"Invalid categories range: {cat_range!r}")
        # titles_from_data=True consumes the first row of data as series titles.
        # If categories start at the same row as data, adjust to skip that header row
        # so category count matches data point count.
        if cat_ref.min_row == data_ref.min_row:
            cat_ref = Reference(
                worksheet=ws,
                min_col=cat_ref.min_col,
                min_row=cat_ref.min_row + 1,
                max_col=cat_ref.max_col,
                max_row=cat_ref.max_row,
            )
        chart.set_categories(cat_ref)

    # Set legend position (normalize aliases like "bottom" → "b")
    legend_pos = op.params.get("legend")
    if legend_pos:
        normalized = _LEGEND_POSITION_MAP.get(legend_pos.lower())
        if normalized is None:
            available = ", ".join(sorted(set(_LEGEND_POSITION_MAP.values())))
            return OpResult(
                success=False,
                message=f"Unknown legend position: {legend_pos!r}. Use: {available}",
            )
        if chart.legend is not None:
            chart.legend.position = normalized
        else:
            from openpyxl.chart.legend import Legend
            chart.legend = Legend()
            chart.legend.position = normalized

    # Parse size WxH (in approximate screen units, convert to cm)
    size_str = op.params.get("size")
    if size_str and "x" in size_str.lower():
        parts = size_str.lower().split("x")
        try:
            w, h = float(parts[0]), float(parts[1])
            chart.width = w / 50  # approximate conversion to cm
            chart.height = h / 50
        except (ValueError, IndexError):
            pass

    # Set anchor position
    anchor = op.params.get("at")
    if anchor:
        chart.anchor = anchor

    # Check for duplicate title
    if title and _find_chart(ws, title) is not None:
        return OpResult(success=False, message=f"Chart with title {title!r} already exists")

    ws.add_chart(chart, anchor or "E5")

    label = title or chart_type
    return OpResult(success=True, message=f"Chart '{label}' added ({chart_type})", prefix="+")


def _chart_series(op: ParsedOp, ctx: SheetsOpContext) -> OpResult:
    """chart series CHART_TITLE data:RANGE [title:"TEXT"]"""
    if len(op.positionals) < 2:
        return OpResult(success=False, message='Usage: chart series CHART_TITLE data:RANGE [title:"TEXT"]')

    chart_title = op.positionals[1]
    ws = ctx.active_sheet
    chart = _find_chart(ws, chart_title)
    if chart is None:
        return OpResult(success=False, message=f"Chart not found: {chart_title!r}")

    data_range = op.params.get("data")
    if not data_range:
        return OpResult(success=False, message="Missing required param: data:RANGE")

    data_ref = _parse_range_to_reference(ws, data_range)
    if data_ref is None:
        return OpResult(success=False, message=f"Invalid data range: {data_range!r}")

    chart.add_data(data_ref, titles_from_data=True)

    series_title = op.params.get("title", data_range)
    return OpResult(success=True, message=f"Series '{series_title}' added to chart '{chart_title}'", prefix="~")


def _chart_axis(op: ParsedOp, ctx: SheetsOpContext) -> OpResult:
    """chart axis CHART_TITLE x|y [title:"TEXT"] [min:N] [max:N] [fmt:FORMAT]"""
    if len(op.positionals) < 3:
        return OpResult(
            success=False,
            message='Usage: chart axis CHART_TITLE x|y [title:"TEXT"] [min:N] [max:N] [fmt:FORMAT]',
        )

    chart_title = op.positionals[1]
    axis_name = op.positionals[2].lower()

    ws = ctx.active_sheet
    chart = _find_chart(ws, chart_title)
    if chart is None:
        return OpResult(success=False, message=f"Chart not found: {chart_title!r}")

    if axis_name not in ("x", "y"):
        return OpResult(success=False, message=f"Invalid axis: {axis_name!r}. Use 'x' or 'y'.")

    axis = chart.x_axis if axis_name == "x" else chart.y_axis

    # Set axis title
    axis_title = op.params.get("title")
    if axis_title:
        axis.title = axis_title

    # Set min
    min_val = op.params.get("min")
    if min_val is not None:
        try:
            axis.scaling.min = float(min_val)
        except (ValueError, TypeError):
            pass

    # Set max
    max_val = op.params.get("max")
    if max_val is not None:
        try:
            axis.scaling.max = float(max_val)
        except (ValueError, TypeError):
            pass

    # Set number format
    fmt = op.params.get("fmt")
    if fmt:
        axis.numFmt = fmt

    parts = []
    if axis_title:
        parts.append(f"title={axis_title!r}")
    if min_val:
        parts.append(f"min={min_val}")
    if max_val:
        parts.append(f"max={max_val}")
    if fmt:
        parts.append(f"fmt={fmt}")

    detail = ", ".join(parts) if parts else "updated"
    return OpResult(success=True, message=f"Chart '{chart_title}' {axis_name}-axis: {detail}", prefix="*")


def _chart_remove(op: ParsedOp, ctx: SheetsOpContext) -> OpResult:
    """chart remove CHART_TITLE"""
    if len(op.positionals) < 2:
        return OpResult(success=False, message="Usage: chart remove CHART_TITLE")

    chart_title = op.positionals[1]
    ws = ctx.active_sheet

    for i, chart in enumerate(ws._charts):
        if _get_chart_title_text(chart) == chart_title:
            ws._charts.pop(i)
            return OpResult(success=True, message=f"Chart '{chart_title}' removed", prefix="-")

    return OpResult(success=False, message=f"Chart not found: {chart_title!r}")


# Sub-command dispatch
_CHART_SUBCMDS: dict[str, callable] = {
    "add": _chart_add,
    "series": _chart_series,
    "axis": _chart_axis,
    "remove": _chart_remove,
}


def op_chart(op: ParsedOp, ctx: SheetsOpContext) -> OpResult:
    """Main chart verb dispatcher."""
    if not op.positionals:
        return OpResult(
            success=False,
            message="Usage: chart add|series|axis|remove ...",
        )

    subcmd = op.positionals[0].lower()
    handler = _CHART_SUBCMDS.get(subcmd)
    if handler is None:
        available = ", ".join(sorted(_CHART_SUBCMDS.keys()))
        return OpResult(success=False, message=f"Unknown chart sub-command: {subcmd!r}. Available: {available}")

    return handler(op, ctx)


HANDLERS: dict[str, callable] = {
    "chart": op_chart,
}



"""SheetIndex — lightweight metadata index for selector resolution.

Per C4 (Lazy Selector Evaluation):
- Tracks data bounds per sheet, active sheet, last modified ranges
- NO per-mutation cell scanning
- Bounds expand on writes, full rebuild only on undo/redo/open
"""

from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from fcp_sheets.model.snapshot import SheetsModel


class SheetIndex:
    """Lazy index tracking sheet metadata for efficient selector resolution."""

    def __init__(self) -> None:
        self._bounds: dict[str, tuple[int, int, int, int]] = {}
        self._active_sheet: str = ""
        self._last_modified: list[tuple[str, str]] = []  # [(sheet, range_str), ...]
        self._named_styles: set[str] = set()

    @property
    def active_sheet(self) -> str:
        return self._active_sheet

    @active_sheet.setter
    def active_sheet(self, name: str) -> None:
        self._active_sheet = name

    def expand_bounds(self, sheet_name: str, row: int, col: int) -> None:
        """Expand data bounds for a sheet after a write operation."""
        if sheet_name in self._bounds:
            mr, mc, xr, xc = self._bounds[sheet_name]
            self._bounds[sheet_name] = (
                min(mr, row), min(mc, col),
                max(xr, row), max(xc, col),
            )
        else:
            self._bounds[sheet_name] = (row, col, row, col)

    def get_bounds(self, sheet_name: str) -> tuple[int, int, int, int] | None:
        """Get data bounds for a sheet: (min_row, min_col, max_row, max_col)."""
        return self._bounds.get(sheet_name)

    def record_modified(self, sheet_name: str, range_str: str) -> None:
        """Record the last modified range (for @recent selector)."""
        self._last_modified.append((sheet_name, range_str))
        # Keep only last 20
        if len(self._last_modified) > 20:
            self._last_modified = self._last_modified[-20:]

    def get_recent(self, count: int = 1) -> list[tuple[str, str]]:
        """Get the last N modified (sheet, range) pairs."""
        return self._last_modified[-count:]

    def remove_sheet(self, sheet_name: str) -> None:
        """Remove a sheet from the index."""
        self._bounds.pop(sheet_name, None)
        self._last_modified = [
            (s, r) for s, r in self._last_modified if s != sheet_name
        ]

    def rename_sheet(self, old_name: str, new_name: str) -> None:
        """Rename a sheet in the index."""
        if old_name in self._bounds:
            self._bounds[new_name] = self._bounds.pop(old_name)
        self._last_modified = [
            (new_name if s == old_name else s, r) for s, r in self._last_modified
        ]
        if self._active_sheet == old_name:
            self._active_sheet = new_name

    def rebuild(self, model: SheetsModel) -> None:
        """Full rebuild from workbook state (after undo/redo/open).

        Uses openpyxl's min/max row/col properties (O(sheets), not O(cells)).
        """
        self._bounds.clear()
        wb = model.wb
        for ws in wb.worksheets:
            if ws.max_row is not None and ws.max_column is not None:
                min_row = ws.min_row or 1
                min_col = ws.min_column or 1
                self._bounds[ws.title] = (min_row, min_col, ws.max_row, ws.max_column)
        self._active_sheet = wb.active.title if wb.active else ""

    def clear(self) -> None:
        """Reset index state."""
        self._bounds.clear()
        self._active_sheet = ""
        self._last_modified.clear()
        self._named_styles.clear()



"""Tests for miscellaneous operations — name, image, link, comment, protect, lock, page-setup."""

from __future__ import annotations

import os
import tempfile

import pytest
from fcp_core import ParsedOp

def _has_pillow() -> bool:
    try:
        import PIL  # noqa: F401
        return True
    except ImportError:
        return False


from fcp_sheets.server.ops_misc import (
    op_comment,
    op_image,
    op_link,
    op_lock,
    op_name,
    op_page_setup,
    op_protect,
    op_unlock,
    op_unprotect,
)
from fcp_sheets.server.resolvers import SheetsOpContext


# ── name ────────────────────────────────────────────────────────────────

class TestNameDefine:
    def test_define_named_range(self, ctx: SheetsOpContext):
        op = ParsedOp(
            verb="name", positionals=["define", "SalesData"],
            params={"range": "$A$1:$D$10"}, raw="name define SalesData range:$A$1:$D$10",
        )
        result = op_name(op, ctx)
        assert result.success
        assert "SalesData" in result.message
        # Verify defined name exists in workbook
        assert "SalesData" in ctx.wb.defined_names

    def test_define_with_scope(self, ctx: SheetsOpContext):
        ctx.wb.create_sheet("Revenue")
        op = ParsedOp(
            verb="name", positionals=["define", "TotalRev"],
            params={"range": "$A$1:$A$100", "scope": "Revenue"},
            raw="name define TotalRev range:$A$1:$A$100 scope:Revenue",
        )
        result = op_name(op, ctx)
        assert result.success
        assert "Revenue" in result.message

    def test_define_missing_range(self, ctx: SheetsOpContext):
        op = ParsedOp(
            verb="name", positionals=["define", "NoRange"],
            raw="name define NoRange",
        )
        result = op_name(op, ctx)
        assert not result.success

    def test_define_missing_name(self, ctx: SheetsOpContext):
        op = ParsedOp(
            verb="name", positionals=["define"],
            raw="name define",
        )
        result = op_name(op, ctx)
        assert not result.success

    def test_define_scope_nonexistent_sheet(self, ctx: SheetsOpContext):
        op = ParsedOp(
            verb="name", positionals=["define", "Test"],
            params={"range": "$A$1", "scope": "Nonexistent"},
            raw="name define Test range:$A$1 scope:Nonexistent",
        )
        result = op_name(op, ctx)
        assert not result.success


class TestNameRemove:
    def test_remove_named_range(self, ctx: SheetsOpContext):
        # First define
        from openpyxl.workbook.defined_name import DefinedName
        defn = DefinedName("TestName", attr_text="'Sheet1'!$A$1:$B$5")
        ctx.wb.defined_names.add(defn)

        op = ParsedOp(
            verb="name", positionals=["remove", "TestName"],
            raw="name remove TestName",
        )
        result = op_name(op, ctx)
        assert result.success

    def test_remove_nonexistent(self, ctx: SheetsOpContext):
        op = ParsedOp(
            verb="name", positionals=["remove", "NoSuchName"],
            raw="name remove NoSuchName",
        )
        result = op_name(op, ctx)
        assert not result.success

    def test_name_unknown_action(self, ctx: SheetsOpContext):
        op = ParsedOp(
            verb="name", positionals=["destroy", "Something"],
            raw="name destroy Something",
        )
        result = op_name(op, ctx)
        assert not result.success

    def test_name_no_args(self, ctx: SheetsOpContext):
        op = ParsedOp(verb="name", positionals=[], raw="name")
        result = op_name(op, ctx)
        assert not result.success


# ── image ───────────────────────────────────────────────────────────────

class TestImage:
    def test_image_file_not_found(self, ctx: SheetsOpContext):
        op = ParsedOp(
            verb="image", positionals=["A1"],
            params={"path": "/nonexistent/image.png"},
            raw="image A1 path:/nonexistent/image.png",
        )
        result = op_image(op, ctx)
        assert not result.success
        assert "not found" in result.message.lower()

    def test_image_missing_path(self, ctx: SheetsOpContext):
        op = ParsedOp(verb="image", positionals=["A1"], raw="image A1")
        result = op_image(op, ctx)
        assert not result.success

    def test_image_no_args(self, ctx: SheetsOpContext):
        op = ParsedOp(verb="image", positionals=[], raw="image")
        result = op_image(op, ctx)
        assert not result.success

    def test_image_invalid_cell(self, ctx: SheetsOpContext):
        op = ParsedOp(
            verb="image", positionals=["INVALID"],
            params={"path": "/tmp/test.png"},
            raw="image INVALID path:/tmp/test.png",
        )
        result = op_image(op, ctx)
        assert not result.success

    @pytest.mark.skipif(
        not _has_pillow(), reason="Pillow not installed"
    )
    def test_image_with_valid_file(self, ctx: SheetsOpContext):
        """Test image insertion with an actual temporary image file."""
        # Create a minimal 1x1 PNG file
        import struct
        import zlib

        def create_minimal_png() -> bytes:
            signature = b'\x89PNG\r\n\x1a\n'
            # IHDR chunk
            ihdr_data = struct.pack('>IIBBBBB', 1, 1, 8, 2, 0, 0, 0)
            ihdr_crc = zlib.crc32(b'IHDR' + ihdr_data)
            ihdr = struct.pack('>I', 13) + b'IHDR' + ihdr_data + struct.pack('>I', ihdr_crc & 0xffffffff)
            # IDAT chunk
            raw_data = b'\x00\xff\x00\x00'  # filter byte + RGB
            compressed = zlib.compress(raw_data)
            idat_crc = zlib.crc32(b'IDAT' + compressed)
            idat = struct.pack('>I', len(compressed)) + b'IDAT' + compressed + struct.pack('>I', idat_crc & 0xffffffff)
            # IEND chunk
            iend_crc = zlib.crc32(b'IEND')
            iend = struct.pack('>I', 0) + b'IEND' + struct.pack('>I', iend_crc & 0xffffffff)
            return signature + ihdr + idat + iend

        with tempfile.NamedTemporaryFile(suffix=".png", delete=False) as f:
            f.write(create_minimal_png())
            tmp_path = f.name

        try:
            op = ParsedOp(
                verb="image", positionals=["A1"],
                params={"path": tmp_path},
                raw=f"image A1 path:{tmp_path}",
            )
            result = op_image(op, ctx)
            assert result.success
        finally:
            os.unlink(tmp_path)

    @pytest.mark.skipif(
        not _has_pillow(), reason="Pillow not installed"
    )
    def test_image_invalid_size_format(self, ctx: SheetsOpContext):
        """Test that invalid size format is rejected."""
        with tempfile.NamedTemporaryFile(suffix=".png", delete=False) as f:
            # Write minimal PNG content
            import struct, zlib
            sig = b'\x89PNG\r\n\x1a\n'
            ihdr_data = struct.pack('>IIBBBBB', 1, 1, 8, 2, 0, 0, 0)
            ihdr_crc = zlib.crc32(b'IHDR' + ihdr_data)
            ihdr = struct.pack('>I', 13) + b'IHDR' + ihdr_data + struct.pack('>I', ihdr_crc & 0xffffffff)
            raw_data = b'\x00\xff\x00\x00'
            compressed = zlib.compress(raw_data)
            idat_crc = zlib.crc32(b'IDAT' + compressed)
            idat = struct.pack('>I', len(compressed)) + b'IDAT' + compressed + struct.pack('>I', idat_crc & 0xffffffff)
            iend_crc = zlib.crc32(b'IEND')
            iend = struct.pack('>I', 0) + b'IEND' + struct.pack('>I', iend_crc & 0xffffffff)
            f.write(sig + ihdr + idat + iend)
            tmp_path = f.name

        try:
            op = ParsedOp(
                verb="image", positionals=["A1"],
                params={"path": tmp_path, "size": "bad"},
                raw=f"image A1 path:{tmp_path} size:bad",
            )
            result = op_image(op, ctx)
            assert not result.success
            assert "size" in result.message.lower()
        finally:
            os.unlink(tmp_path)


# ── link ────────────────────────────────────────────────────────────────

class TestLink:
    def test_link_url(self, ctx: SheetsOpContext):
        op = ParsedOp(
            verb="link", positionals=["A1"],
            params={"url": "https://example.com"},
            raw="link A1 url:https://example.com",
        )
        result = op_link(op, ctx)
        assert result.success
        cell = ctx.active_sheet.cell(row=1, column=1)
        assert cell.hyperlink is not None
        assert cell.value == "https://example.com"

    def test_link_url_with_text(self, ctx: SheetsOpContext):
        op = ParsedOp(
            verb="link", positionals=["B2"],
            params={"url": "https://example.com", "text": "Click here"},
            raw='link B2 url:https://example.com text:"Click here"',
        )
        result = op_link(op, ctx)
        assert result.success
        cell = ctx.active_sheet.cell(row=2, column=2)
        assert cell.value == "Click here"

    def test_link_internal_sheet(self, ctx: SheetsOpContext):
        ctx.wb.create_sheet("Data")
        op = ParsedOp(
            verb="link", positionals=["A1"],
            params={"sheet": "Data!A1"},
            raw="link A1 sheet:Data!A1",
        )
        result = op_link(op, ctx)
        assert result.success
        cell = ctx.active_sheet.cell(row=1, column=1)
        assert cell.hyperlink is not None

    def test_link_internal_sheet_no_cell(self, ctx: SheetsOpContext):
        """Link to a sheet without specifying a cell defaults to A1."""
        ctx.wb.create_sheet("Summary")
        op = ParsedOp(
            verb="link", positionals=["A1"],
            params={"sheet": "Summary"},
            raw="link A1 sheet:Summary",
        )
        result = op_link(op, ctx)
        assert result.success

    def test_link_internal_nonexistent_sheet(self, ctx: SheetsOpContext):
        op = ParsedOp(
            verb="link", positionals=["A1"],
            params={"sheet": "Nonexistent!A1"},
            raw="link A1 sheet:Nonexistent!A1",
        )
        result = op_link(op, ctx)
        assert not result.success

    def test_link_off(self, ctx: SheetsOpContext):
        # First add a link
        cell = ctx.active_sheet.cell(row=1, column=1)
        cell.hyperlink = "https://example.com"
        cell.value = "Link"

        op = ParsedOp(verb="link", positionals=["off", "A1"], raw="link off A1")
        result = op_link(op, ctx)
        assert result.success
        assert ctx.active_sheet.cell(row=1, column=1).hyperlink is None

    def test_link_off_missing_cell(self, ctx: SheetsOpContext):
        op = ParsedOp(verb="link", positionals=["off"], raw="link off")
        result = op_link(op, ctx)
        assert not result.success

    def test_link_no_args(self, ctx: SheetsOpContext):
        op = ParsedOp(verb="link", positionals=[], raw="link")
        result = op_link(op, ctx)
        assert not result.success

    def test_link_missing_url_and_sheet(self, ctx: SheetsOpContext):
        op = ParsedOp(verb="link", positionals=["A1"], raw="link A1")
        result = op_link(op, ctx)
        assert not result.success


# ── comment ─────────────────────────────────────────────────────────────

class TestComment:
    def test_add_comment(self, ctx: SheetsOpContext):
        op = ParsedOp(
            verb="comment", positionals=["A1", "This is a note"],
            raw='comment A1 "This is a note"',
        )
        result = op_comment(op, ctx)
        assert result.success
        cell = ctx.active_sheet.cell(row=1, column=1)
        assert cell.comment is not None
        assert cell.comment.text == "This is a note"

    def test_remove_comment(self, ctx: SheetsOpContext):
        from openpyxl.comments import Comment
        cell = ctx.active_sheet.cell(row=1, column=1)
        cell.comment = Comment("Old note", "Author")

        op = ParsedOp(verb="comment", positionals=["off", "A1"], raw="comment off A1")
        result = op_comment(op, ctx)
        assert result.success
        assert ctx.active_sheet.cell(row=1, column=1).comment is None

    def test_comment_off_missing_cell(self, ctx: SheetsOpContext):
        op = ParsedOp(verb="comment", positionals=["off"], raw="comment off")
        result = op_comment(op, ctx)
        assert not result.success

    def test_comment_missing_text(self, ctx: SheetsOpContext):
        op = ParsedOp(verb="comment", positionals=["A1"], raw="comment A1")
        result = op_comment(op, ctx)
        assert not result.success

    def test_comment_no_args(self, ctx: SheetsOpContext):
        op = ParsedOp(verb="comment", positionals=[], raw="comment")
        result = op_comment(op, ctx)
        assert not result.success

    def test_comment_invalid_cell(self, ctx: SheetsOpContext):
        op = ParsedOp(
            verb="comment", positionals=["INVALID", "Note"],
            raw='comment INVALID "Note"',
        )
        result = op_comment(op, ctx)
        assert not result.success


# ── protect / unprotect ─────────────────────────────────────────────────

class TestProtect:
    def test_protect(self, ctx: SheetsOpContext):
        op = ParsedOp(verb="protect", positionals=[], raw="protect")
        result = op_protect(op, ctx)
        assert result.success
        assert ctx.active_sheet.protection.sheet is True

    def test_protect_with_password(self, ctx: SheetsOpContext):
        op = ParsedOp(
            verb="protect", positionals=[],
            params={"password": "secret123"},
            raw="protect password:secret123",
        )
        result = op_protect(op, ctx)
        assert result.success
        assert ctx.active_sheet.protection.sheet is True
        assert "password" in result.message.lower()

    def test_unprotect(self, ctx: SheetsOpContext):
        ctx.active_sheet.protection.sheet = True
        op = ParsedOp(verb="unprotect", positionals=[], raw="unprotect")
        result = op_unprotect(op, ctx)
        assert result.success
        assert ctx.active_sheet.protection.sheet is False


# ── lock / unlock ───────────────────────────────────────────────────────

class TestLockUnlock:
    def test_lock_range(self, ctx: SheetsOpContext):
        op = ParsedOp(verb="lock", positionals=["A1:B2"], raw="lock A1:B2")
        result = op_lock(op, ctx)
        assert result.success
        for r in range(1, 3):
            for c in range(1, 3):
                assert ctx.active_sheet.cell(row=r, column=c).protection.locked is True

    def test_lock_single_cell(self, ctx: SheetsOpContext):
        op = ParsedOp(verb="lock", positionals=["C3"], raw="lock C3")
        result = op_lock(op, ctx)
        assert result.success
        assert ctx.active_sheet.cell(row=3, column=3).protection.locked is True

    def test_unlock_range(self, ctx: SheetsOpContext):
        # Lock first
        from openpyxl.styles import Protection
        for r in range(1, 3):
            for c in range(1, 3):
                ctx.active_sheet.cell(row=r, column=c).protection = Protection(locked=True)

        op = ParsedOp(verb="unlock", positionals=["A1:B2"], raw="unlock A1:B2")
        result = op_unlock(op, ctx)
        assert result.success
        for r in range(1, 3):
            for c in range(1, 3):
                assert ctx.active_sheet.cell(row=r, column=c).protection.locked is False

    def test_lock_no_args(self, ctx: SheetsOpContext):
        op = ParsedOp(verb="lock", positionals=[], raw="lock")
        result = op_lock(op, ctx)
        assert not result.success

    def test_unlock_no_args(self, ctx: SheetsOpContext):
        op = ParsedOp(verb="unlock", positionals=[], raw="unlock")
        result = op_unlock(op, ctx)
        assert not result.success

    def test_lock_invalid_range(self, ctx: SheetsOpContext):
        op = ParsedOp(verb="lock", positionals=["INVALID"], raw="lock INVALID")
        result = op_lock(op, ctx)
        assert not result.success


# ── page-setup ──────────────────────────────────────────────────────────

class TestPageSetup:
    def test_orientation_landscape(self, ctx: SheetsOpContext):
        op = ParsedOp(
            verb="page-setup", positionals=[],
            params={"orient": "landscape"},
            raw="page-setup orient:landscape",
        )
        result = op_page_setup(op, ctx)
        assert result.success
        assert ctx.active_sheet.page_setup.orientation == "landscape"

    def test_orientation_portrait(self, ctx: SheetsOpContext):
        op = ParsedOp(
            verb="page-setup", positionals=[],
            params={"orient": "portrait"},
            raw="page-setup orient:portrait",
        )
        result = op_page_setup(op, ctx)
        assert result.success
        assert ctx.active_sheet.page_setup.orientation == "portrait"

    def test_invalid_orientation(self, ctx: SheetsOpContext):
        op = ParsedOp(
            verb="page-setup", positionals=[],
            params={"orient": "sideways"},
            raw="page-setup orient:sideways",
        )
        result = op_page_setup(op, ctx)
        assert not result.success

    def test_paper_size(self, ctx: SheetsOpContext):
        op = ParsedOp(
            verb="page-setup", positionals=[],
            params={"paper": "a4"},
            raw="page-setup paper:a4",
        )
        result = op_page_setup(op, ctx)
        assert result.success
        assert ctx.active_sheet.page_setup.paperSize == 9  # A4

    def test_unknown_paper_size(self, ctx: SheetsOpContext):
        op = ParsedOp(
            verb="page-setup", positionals=[],
            params={"paper": "napkin"},
            raw="page-setup paper:napkin",
        )
        result = op_page_setup(op, ctx)
        assert not result.success

    def test_margins(self, ctx: SheetsOpContext):
        op = ParsedOp(
            verb="page-setup", positionals=[],
            params={"margins": "1.0,0.75,1.0,0.75"},
            raw="page-setup margins:1.0,0.75,1.0,0.75",
        )
        result = op_page_setup(op, ctx)
        assert result.success
        margins = ctx.active_sheet.page_margins
        assert margins.top == 1.0
        assert margins.right == 0.75
        assert margins.bottom == 1.0
        assert margins.left == 0.75

    def test_margins_wrong_count(self, ctx: SheetsOpContext):
        op = ParsedOp(
            verb="page-setup", positionals=[],
            params={"margins": "1.0,0.75"},
            raw="page-setup margins:1.0,0.75",
        )
        result = op_page_setup(op, ctx)
        assert not result.success

    def test_print_area(self, ctx: SheetsOpContext):
        op = ParsedOp(
            verb="page-setup", positionals=[],
            params={"print-area": "A1:D20"},
            raw="page-setup print-area:A1:D20",
        )
        result = op_page_setup(op, ctx)
        assert result.success
        # openpyxl normalizes print_area to include sheet name and absolute refs
        assert "A" in ctx.active_sheet.print_area
        assert "D" in ctx.active_sheet.print_area
        assert "20" in ctx.active_sheet.print_area

    def test_fit_to_page(self, ctx: SheetsOpContext):
        op = ParsedOp(
            verb="page-setup", positionals=[],
            params={"fit-width": "1", "fit-height": "0"},
            raw="page-setup fit-width:1 fit-height:0",
        )
        result = op_page_setup(op, ctx)
        assert result.success
        assert ctx.active_sheet.page_setup.fitToWidth == 1
        assert ctx.active_sheet.page_setup.fitToHeight == 0

    def test_gridlines_flag(self, ctx: SheetsOpContext):
        op = ParsedOp(
            verb="page-setup", positionals=["gridlines"],
            raw="page-setup gridlines",
        )
        result = op_page_setup(op, ctx)
        assert result.success
        assert ctx.active_sheet.print_options.gridLines is True

    def test_center_flags(self, ctx: SheetsOpContext):
        op = ParsedOp(
            verb="page-setup", positionals=["center-h", "center-v"],
            raw="page-setup center-h center-v",
        )
        result = op_page_setup(op, ctx)
        assert result.success
        assert ctx.active_sheet.print_options.horizontalCentered is True
        assert ctx.active_sheet.print_options.verticalCentered is True

    def test_no_changes(self, ctx: SheetsOpContext):
        op = ParsedOp(verb="page-setup", positionals=[], raw="page-setup")
        result = op_page_setup(op, ctx)
        assert result.success
        assert "no" in result.message.lower()

    def test_multiple_settings(self, ctx: SheetsOpContext):
        op = ParsedOp(
            verb="page-setup", positionals=["gridlines"],
            params={"orient": "landscape", "paper": "letter", "print-area": "A1:Z50"},
            raw="page-setup orient:landscape paper:letter print-area:A1:Z50 gridlines",
        )
        result = op_page_setup(op, ctx)
        assert result.success
        ws = ctx.active_sheet
        assert ws.page_setup.orientation == "landscape"
        assert ws.page_setup.paperSize == 1
        assert "A" in ws.print_area and "Z" in ws.print_area and "50" in ws.print_area
        assert ws.print_options.gridLines is True

    def test_header_footer(self, ctx: SheetsOpContext):
        op = ParsedOp(
            verb="page-setup", positionals=[],
            params={"header": "My Report", "footer": "Page &P"},
            raw='page-setup header:"My Report" footer:"Page &P"',
        )
        result = op_page_setup(op, ctx)
        assert result.success
        assert ctx.active_sheet.oddHeader.center.text == "My Report"
        assert ctx.active_sheet.oddFooter.center.text == "Page &P"

    def test_print_title_rows(self, ctx: SheetsOpContext):
        op = ParsedOp(
            verb="page-setup", positionals=[],
            params={"print-title-rows": "1:2"},
            raw="page-setup print-title-rows:1:2",
        )
        result = op_page_setup(op, ctx)
        assert result.success
        assert ctx.active_sheet.print_title_rows == "$1:$2"
        assert "print-title-rows" in result.message

    def test_print_title_cols(self, ctx: SheetsOpContext):
        op = ParsedOp(
            verb="page-setup", positionals=[],
            params={"print-title-cols": "A:B"},
            raw="page-setup print-title-cols:A:B",
        )
        result = op_page_setup(op, ctx)
        assert result.success
        assert ctx.active_sheet.print_title_cols == "$A:$B"
        assert "print-title-cols" in result.message

    def test_print_title_rows_and_cols(self, ctx: SheetsOpContext):
        op = ParsedOp(
            verb="page-setup", positionals=[],
            params={"print-title-rows": "1:1", "print-title-cols": "A:A"},
            raw="page-setup print-title-rows:1:1 print-title-cols:A:A",
        )
        result = op_page_setup(op, ctx)
        assert result.success
        assert ctx.active_sheet.print_title_rows == "$1:$1"
        assert ctx.active_sheet.print_title_cols == "$A:$A"



"""Query handlers — read-only inspection of workbook state.

MVP queries: plan/map, stats, status, history.
Extended queries (Wave 3): describe, peek, list, find.
"""

from __future__ import annotations

import re
from datetime import datetime

from fcp_sheets.model.index import SheetIndex
from fcp_sheets.model.refs import (
    index_to_col,
    col_to_index,
    parse_cell_ref,
    parse_range_ref,
    CellRef,
    RangeRef,
)
from fcp_sheets.model.snapshot import SheetsModel
from fcp_sheets.server.formatter import (
    format_range,
    format_cell_addr,
    truncate_list,
    format_cell_value,
    format_value_type,
    format_font,
    format_fill,
    format_alignment,
    format_border,
    format_table_row,
)


def dispatch_query(query: str, model: SheetsModel, index: SheetIndex) -> str:
    """Route a query string to the appropriate handler."""
    query = query.strip()
    parts = query.split(None, 1)
    command = parts[0].lower() if parts else ""
    args = parts[1] if len(parts) > 1 else ""

    handlers = {
        "plan": _query_plan,
        "map": _query_plan,
        "stats": _query_stats,
        "status": _query_status,
        "history": _query_history,
        "describe": _query_describe,
        "peek": _query_peek,
        "list": _query_list,
        "find": _query_find,
    }

    handler = handlers.get(command)
    if handler is None:
        return (
            f"! Unknown query: {command!r}\n"
            "  try: plan, stats, status, history, describe, peek, list, find"
        )

    return handler(args, model, index)


# ---------------------------------------------------------------------------
# plan / map — primary overview
# ---------------------------------------------------------------------------

def _query_plan(args: str, model: SheetsModel, index: SheetIndex) -> str:
    """Show workbook topology — the most important query."""
    wb = model.wb
    lines: list[str] = []

    # Header
    saved = f", saved: {model.file_path}" if model.file_path else ", unsaved"
    lines.append(f'Workbook: "{model.title}" ({len(wb.sheetnames)} sheets{saved})')

    for ws in wb.worksheets:
        sheet_name = ws.title
        active = " [active]" if ws == wb.active else ""
        hidden = " [hidden]" if ws.sheet_state == "hidden" else ""
        lines.append(f"\n  Sheet: {sheet_name}{active}{hidden}")

        bounds = index.get_bounds(sheet_name)
        if bounds:
            min_row, min_col, max_row, max_col = bounds
            range_str = format_range(min_row, min_col, max_row, max_col)

            # Data bounds + freeze + filter
            meta_parts = [f"data: {range_str}"]
            if ws.freeze_panes:
                meta_parts.append(f"frozen: {ws.freeze_panes}")
            if ws.auto_filter and ws.auto_filter.ref:
                meta_parts.append(f"filter: {ws.auto_filter.ref}")
            lines.append(f"    {' | '.join(meta_parts)}")

            # Column headers (first row)
            cols = []
            for col in range(min_col, min(max_col + 1, min_col + 8)):
                cell = ws.cell(row=min_row, column=col)
                val = cell.value
                if val is not None:
                    cols.append(f"{index_to_col(col)}:{val}")
            if cols:
                lines.append(f"    cols: {truncate_list(cols)}")

            # Formula patterns
            formula_patterns = _detect_formula_patterns(ws, bounds)
            if formula_patterns:
                lines.append(f"    formulas: {' | '.join(formula_patterns)}")

            # Tables
            if ws.tables:
                for table_name in ws.tables:
                    table = ws.tables[table_name]
                    lines.append(f"    table: {table_name} {table.ref}")

            # Charts
            if ws._charts:
                for chart in ws._charts:
                    chart_title = chart.title or "Untitled"
                    chart_type = type(chart).__name__
                    lines.append(f"    chart: {chart_type} \"{chart_title}\"")

            # Conditional formatting
            if ws.conditional_formatting:
                cf_count = len(list(ws.conditional_formatting))
                if cf_count > 0:
                    lines.append(f"    cond-fmt: {cf_count} rule(s)")

            # Next empty row
            lines.append(f"    next-empty: row:{max_row + 1}")
        else:
            lines.append("    (empty)")

    return "\n".join(lines)


# ---------------------------------------------------------------------------
# stats
# ---------------------------------------------------------------------------

def _query_stats(args: str, model: SheetsModel, index: SheetIndex) -> str:
    """Quick summary of workbook contents."""
    wb = model.wb
    total_data = 0
    total_formula = 0
    total_charts = 0
    total_tables = 0
    total_merged = 0
    total_cond_fmt = 0

    for ws in wb.worksheets:
        bounds = index.get_bounds(ws.title)
        if bounds:
            min_row, min_col, max_row, max_col = bounds
            for row in ws.iter_rows(
                min_row=min_row, max_row=max_row,
                min_col=min_col, max_col=max_col,
            ):
                for cell in row:
                    if cell.value is not None:
                        if isinstance(cell.value, str) and cell.value.startswith("="):
                            total_formula += 1
                        else:
                            total_data += 1

        total_charts += len(ws._charts)
        total_tables += len(ws.tables)
        total_merged += len(ws.merged_cells.ranges)
        total_cond_fmt += len(list(ws.conditional_formatting))

    named_ranges = len(list(wb.defined_names)) if wb.defined_names else 0

    lines = [
        f'Workbook: "{model.title}"',
        f"  Sheets: {len(wb.sheetnames)} ({', '.join(wb.sheetnames)})",
        f"  Data cells: {total_data:,} | Formula cells: {total_formula:,}",
        f"  Tables: {total_tables} | Charts: {total_charts} | Named ranges: {named_ranges}",
        f"  Merged regions: {total_merged} | Conditional formats: {total_cond_fmt}",
    ]
    return "\n".join(lines)


# ---------------------------------------------------------------------------
# status
# ---------------------------------------------------------------------------

def _query_status(args: str, model: SheetsModel, index: SheetIndex) -> str:
    """Session status."""
    saved = model.file_path or "unsaved"
    active = model.wb.active.title if model.wb.active else "none"
    return (
        f'Session: "{model.title}"\n'
        f"  File: {saved}\n"
        f"  Sheets: {len(model.wb.sheetnames)}\n"
        f"  Active: {active}"
    )


# ---------------------------------------------------------------------------
# history
# ---------------------------------------------------------------------------

def _query_history(args: str, model: SheetsModel, index: SheetIndex) -> str:
    """Show recent operations — managed by session layer."""
    return "History managed by session layer. Use sheets_session for undo/redo."


# ---------------------------------------------------------------------------
# describe — detailed view of sheet, cell, or range
# ---------------------------------------------------------------------------

def _query_describe(args: str, model: SheetsModel, index: SheetIndex) -> str:
    """Describe a sheet, cell, or range."""
    args = args.strip()
    if not args:
        return "! Usage: describe SHEET | describe CELL | describe RANGE"

    wb = model.wb

    # Check if it's a sheet name (exact match first, then case-insensitive)
    if args in wb.sheetnames:
        return _describe_sheet(args, model, index)
    for name in wb.sheetnames:
        if name.lower() == args.lower():
            return _describe_sheet(name, model, index)

    # Check if it's a cell reference (possibly with sheet prefix)
    cell_ref = parse_cell_ref(args)
    if cell_ref:
        return _describe_cell(cell_ref, model, index)

    # Check if it's a range reference
    range_ref = parse_range_ref(args)
    if isinstance(range_ref, RangeRef):
        return _describe_range(range_ref, model, index)

    return f"! Cannot resolve: {args!r}. Provide a sheet name, cell (A1), or range (A1:D10)."


def _describe_sheet(sheet_name: str, model: SheetsModel, index: SheetIndex) -> str:
    """Detailed sheet description."""
    wb = model.wb
    ws = wb[sheet_name]
    lines: list[str] = [f"Sheet: {sheet_name}"]

    bounds = index.get_bounds(sheet_name)
    if not bounds:
        lines.append("  (empty)")
        return "\n".join(lines)

    min_row, min_col, max_row, max_col = bounds
    range_str = format_range(min_row, min_col, max_row, max_col)

    # Meta line
    meta_parts = [f"data: {range_str}"]
    if ws.freeze_panes:
        meta_parts.append(f"frozen: {ws.freeze_panes}")
    if ws.auto_filter and ws.auto_filter.ref:
        meta_parts.append(f"filter: {ws.auto_filter.ref}")
    lines.append(f"  {' | '.join(meta_parts)}")

    # Column analysis
    lines.append("  columns:")
    for col in range(min_col, max_col + 1):
        col_letter = index_to_col(col)
        header_val = ws.cell(row=min_row, column=col).value
        header = str(header_val) if header_val is not None else ""

        # Scan column for type info (skip header row)
        values: list = []
        formulas: list[str] = []
        unique_texts: set[str] = set()
        nums: list[float] = []
        data_count = 0

        for row in range(min_row + 1, max_row + 1):
            val = ws.cell(row=row, column=col).value
            if val is None:
                continue
            data_count += 1
            if isinstance(val, str) and val.startswith("="):
                formulas.append(val)
            elif isinstance(val, (int, float)) and not isinstance(val, bool):
                nums.append(float(val))
            elif isinstance(val, str):
                unique_texts.add(val)
            values.append(val)

        # Determine dominant type
        if formulas and len(formulas) >= len(values) / 2:
            col_type = "formula"
        elif nums and len(nums) >= len(values) / 2:
            col_type = "number"
        else:
            col_type = "text"

        # Build column line
        header_display = header[:12] if header else "(no header)"
        col_line = f"    {col_letter:<3}{header_display:<14}{col_type:<9}{data_count} values"

        if col_type == "text" and unique_texts:
            col_line += f"   unique:{len(unique_texts)}"
        elif col_type == "number" and nums:
            mn = int(min(nums)) if min(nums) == int(min(nums)) else min(nums)
            mx = int(max(nums)) if max(nums) == int(max(nums)) else max(nums)
            col_line += f"   range:{mn}..{mx}"
        elif col_type == "formula" and formulas:
            # Show pattern from first formula
            pattern = re.sub(r"\d+", "N", formulas[0])
            col_line += f"   pattern:{pattern}"
            if nums:
                mn = int(min(nums)) if min(nums) == int(min(nums)) else min(nums)
                mx = int(max(nums)) if max(nums) == int(max(nums)) else max(nums)
                col_line += f"  range:{mn}..{mx}"

        lines.append(col_line)

    # Sample data (rows 2-4, i.e. first 3 data rows after header)
    sample_start = min_row + 1
    sample_end = min(min_row + 3, max_row)
    if sample_start <= max_row:
        lines.append(f"  sample (rows {sample_start}-{sample_end}):")
        # Build header
        headers = []
        for col in range(min_col, max_col + 1):
            val = ws.cell(row=min_row, column=col).value
            headers.append(str(val) if val is not None else index_to_col(col))

        col_widths = [max(8, len(h)) for h in headers]
        header_line = "    | " + " | ".join(h.ljust(w) for h, w in zip(headers, col_widths)) + " |"
        lines.append(header_line)

        for row in range(sample_start, sample_end + 1):
            row_vals = []
            for col in range(min_col, max_col + 1):
                val = ws.cell(row=row, column=col).value
                row_vals.append(_compact_value(val))
            row_line = "    | " + " | ".join(v.ljust(w) for v, w in zip(row_vals, col_widths)) + " |"
            lines.append(row_line)

    # Formula groupings
    formula_groups = _get_formula_groups(ws, bounds)
    if formula_groups:
        lines.append("  formulas:")
        for group in formula_groups[:5]:
            lines.append(f"    {group}")

    # Merged regions
    merged = list(ws.merged_cells.ranges)
    if merged:
        merged_strs = [str(m) for m in merged]
        lines.append(f"  merged: {truncate_list(merged_strs, 5)}")
    else:
        lines.append("  merged: (none)")

    # Conditional formatting
    cf_rules = list(ws.conditional_formatting)
    if cf_rules:
        for cf in cf_rules[:5]:
            # cf is a ConditionalFormattingList entry
            range_str_cf = str(cf)
            for rule in cf.rules:
                rule_type = _describe_cf_rule(rule)
                lines.append(f"  cond-fmt: {range_str_cf} {rule_type}")
        if len(cf_rules) > 5:
            lines.append(f"  ... +{len(cf_rules) - 5} more rules")
    else:
        lines.append("  cond-fmt: (none)")

    # Tables
    if ws.tables:
        for table_name in ws.tables:
            table = ws.tables[table_name]
            style_name = ""
            if table.tableStyleInfo and table.tableStyleInfo.name:
                style_name = f" style:{table.tableStyleInfo.name}"
            banded = ""
            if table.tableStyleInfo:
                flags = []
                if table.tableStyleInfo.showRowStripes:
                    flags.append("banded-rows")
                if table.tableStyleInfo.showColumnStripes:
                    flags.append("banded-cols")
                if flags:
                    banded = " " + " ".join(flags)
            lines.append(f"  table: {table_name} {table.ref}{style_name}{banded}")
    else:
        lines.append("  tables: (none)")

    return "\n".join(lines)


def _describe_cell(ref: CellRef, model: SheetsModel, index: SheetIndex) -> str:
    """Detailed cell description."""
    wb = model.wb

    # Determine sheet
    if ref.sheet and ref.sheet in wb.sheetnames:
        ws = wb[ref.sheet]
        sheet_name = ref.sheet
    else:
        ws = wb.active
        sheet_name = ws.title

    cell = ws.cell(row=ref.row, column=ref.col)
    addr = format_cell_addr(ref.col, ref.row)
    lines: list[str] = [f"Cell {addr} on {sheet_name}"]

    # Value
    val = cell.value
    val_type = format_value_type(val)
    val_display = format_cell_value(val)
    lines.append(f"  value: {val_display} ({val_type})")

    # Font
    lines.append(f"  font: {format_font(cell.font)}")

    # Fill
    lines.append(f"  fill: {format_fill(cell.fill)}")

    # Alignment
    lines.append(f"  alignment: {format_alignment(cell.alignment)}")

    # Border
    lines.append(f"  border: {format_border(cell.border)}")

    # Number format
    if cell.number_format and cell.number_format != "General":
        lines.append(f"  number-format: {cell.number_format}")

    # Part-of table?
    in_table = None
    for table_name in ws.tables:
        table = ws.tables[table_name]
        rr = parse_range_ref(table.ref)
        if isinstance(rr, RangeRef):
            if (rr.start.row <= ref.row <= rr.end.row and
                    rr.start.col <= ref.col <= rr.end.col):
                in_table = table_name
                break

    # Merged?
    in_merged = False
    for merged_range in ws.merged_cells.ranges:
        if cell.coordinate in merged_range:
            in_merged = True
            break

    parts = []
    if in_table:
        parts.append(f"table {in_table}")
    if in_merged:
        parts.append("merged: yes")
    else:
        parts.append("merged: no")
    lines.append(f"  part-of: {', '.join(parts)}")

    return "\n".join(lines)


def _describe_range(ref: RangeRef, model: SheetsModel, index: SheetIndex) -> str:
    """Range summary description."""
    wb = model.wb

    if ref.sheet and ref.sheet in wb.sheetnames:
        ws = wb[ref.sheet]
        sheet_name = ref.sheet
    else:
        ws = wb.active
        sheet_name = ws.title

    start_addr = format_cell_addr(ref.start.col, ref.start.row)
    end_addr = format_cell_addr(ref.end.col, ref.end.row)
    lines: list[str] = [f"Range {start_addr}:{end_addr} on {sheet_name}"]

    # Scan the range
    total = 0
    non_empty = 0
    type_counts: dict[str, int] = {}
    nums: list[float] = []

    for row in range(ref.start.row, ref.end.row + 1):
        for col in range(ref.start.col, ref.end.col + 1):
            total += 1
            val = ws.cell(row=row, column=col).value
            vtype = format_value_type(val)
            type_counts[vtype] = type_counts.get(vtype, 0) + 1
            if val is not None:
                non_empty += 1
                if isinstance(val, (int, float)) and not isinstance(val, bool):
                    nums.append(float(val))

    rows = ref.end.row - ref.start.row + 1
    cols = ref.end.col - ref.start.col + 1
    lines.append(f"  size: {rows} rows x {cols} cols ({total} cells, {non_empty} non-empty)")

    # Type breakdown
    type_parts = []
    for t in ["text", "number", "formula", "date", "empty"]:
        if t in type_counts and type_counts[t] > 0:
            type_parts.append(f"{t}:{type_counts[t]}")
    if type_parts:
        lines.append(f"  types: {', '.join(type_parts)}")

    # Number range
    if nums:
        mn = int(min(nums)) if min(nums) == int(min(nums)) else min(nums)
        mx = int(max(nums)) if max(nums) == int(max(nums)) else max(nums)
        lines.append(f"  number-range: {mn}..{mx}")

    return "\n".join(lines)


# ---------------------------------------------------------------------------
# peek — token-efficient data view
# ---------------------------------------------------------------------------

_PEEK_ROW_CAP = 50

def _query_peek(args: str, model: SheetsModel, index: SheetIndex) -> str:
    """Show actual cell data for a range."""
    args = args.strip()
    if not args:
        return "! Usage: peek RANGE (e.g. peek A1:E10)"

    wb = model.wb

    # Parse range ref (may include sheet prefix)
    range_ref = parse_range_ref(args)
    if not isinstance(range_ref, RangeRef):
        # Maybe single cell — treat as 1x1 range
        cell_ref = parse_cell_ref(args)
        if cell_ref:
            range_ref = RangeRef(
                start=CellRef(col=cell_ref.col, row=cell_ref.row),
                end=CellRef(col=cell_ref.col, row=cell_ref.row),
                sheet=cell_ref.sheet,
            )
        else:
            # Maybe it's a sheet name — peek entire data bounds
            if args in wb.sheetnames:
                bounds = index.get_bounds(args)
                if not bounds:
                    return f"Sheet {args} is empty"
                min_row, min_col, max_row, max_col = bounds
                range_ref = RangeRef(
                    start=CellRef(col=min_col, row=min_row),
                    end=CellRef(col=max_col, row=max_row),
                    sheet=args,
                )
            else:
                return f"! Invalid range: {args!r}"

    # Determine sheet
    if range_ref.sheet and range_ref.sheet in wb.sheetnames:
        ws = wb[range_ref.sheet]
        sheet_name = range_ref.sheet
    else:
        ws = wb.active
        sheet_name = ws.title

    start_row = range_ref.start.row
    end_row = range_ref.end.row
    start_col = range_ref.start.col
    end_col = range_ref.end.col
    num_cols = end_col - start_col + 1
    num_rows = end_row - start_row + 1

    # Enforce row cap
    capped = False
    actual_end_row = end_row
    if num_rows > _PEEK_ROW_CAP:
        actual_end_row = start_row + _PEEK_ROW_CAP - 1
        capped = True

    start_addr = format_cell_addr(start_col, start_row)
    end_addr = format_cell_addr(end_col, end_row)

    lines: list[str] = [f"{start_addr}:{end_addr} on {sheet_name}"]

    if num_cols < 12:
        # Narrow mode — compact table
        # Build all row data first to compute column widths
        all_rows: list[list[str]] = []
        for row in range(start_row, actual_end_row + 1):
            row_vals: list[str] = []
            for col in range(start_col, end_col + 1):
                val = ws.cell(row=row, column=col).value
                row_vals.append(_compact_value(val))
            all_rows.append(row_vals)

        if not all_rows:
            lines.append("  (empty range)")
            return "\n".join(lines)

        # Compute column widths
        col_widths = [0] * num_cols
        for row_vals in all_rows:
            for i, v in enumerate(row_vals):
                col_widths[i] = max(col_widths[i], len(v))

        # Cap widths
        col_widths = [min(w, 20) for w in col_widths]

        # Output rows
        for row_vals in all_rows:
            cells = [v[:20].ljust(col_widths[i]) for i, v in enumerate(row_vals)]
            lines.append("  " + "|".join(cells))
    else:
        # Wide mode — vertical per row
        for row in range(start_row, actual_end_row + 1):
            lines.append(f"  Row {row} on {sheet_name}")
            shown = 0
            hidden = 0
            for col in range(start_col, end_col + 1):
                val = ws.cell(row=row, column=col).value
                if shown < 10:
                    col_letter = index_to_col(col)
                    lines.append(f"    {col_letter} = {_compact_value(val)}")
                    shown += 1
                else:
                    hidden += 1
            if hidden > 0:
                lines.append(f"    ...+{hidden} more cols")

    if capped:
        remaining = num_rows - _PEEK_ROW_CAP
        lines.append(f"  ... +{remaining} more rows (capped at {_PEEK_ROW_CAP})")

    return "\n".join(lines)


# ---------------------------------------------------------------------------
# list — sub-command dispatcher
# ---------------------------------------------------------------------------

def _query_list(args: str, model: SheetsModel, index: SheetIndex) -> str:
    """List sub-command dispatcher."""
    args = args.strip().lower()

    subcmds = {
        "sheets": _list_sheets,
        "charts": _list_charts,
        "formulas": _list_formulas,
        "styles": _list_styles,
        "names": _list_names,
        "tables": _list_tables,
    }

    handler = subcmds.get(args)
    if handler is None:
        available = ", ".join(sorted(subcmds.keys()))
        return f"! Usage: list {available}"

    return handler(model, index)


def _list_sheets(model: SheetsModel, index: SheetIndex) -> str:
    """List sheets with basic stats."""
    wb = model.wb
    lines: list[str] = [f"Sheets ({len(wb.sheetnames)}):"]

    for ws in wb.worksheets:
        sheet_name = ws.title
        bounds = index.get_bounds(sheet_name)
        if bounds:
            min_row, min_col, max_row, max_col = bounds
            rows = max_row - min_row + 1
            cols = max_col - min_col + 1

            # Count non-empty cells
            data_cells = 0
            for row in ws.iter_rows(
                min_row=min_row, max_row=max_row,
                min_col=min_col, max_col=max_col,
            ):
                for cell in row:
                    if cell.value is not None:
                        data_cells += 1

            active = " [active]" if ws == wb.active else ""
            hidden = " [hidden]" if ws.sheet_state == "hidden" else ""
            lines.append(
                f"  {sheet_name}{active}{hidden}: "
                f"{rows} rows, {cols} cols, {data_cells} data cells"
            )
        else:
            active = " [active]" if ws == wb.active else ""
            hidden = " [hidden]" if ws.sheet_state == "hidden" else ""
            lines.append(f"  {sheet_name}{active}{hidden}: (empty)")

    return "\n".join(lines)


def _list_charts(model: SheetsModel, index: SheetIndex) -> str:
    """List all charts across all sheets."""
    wb = model.wb
    charts_found: list[str] = []

    for ws in wb.worksheets:
        for chart in ws._charts:
            chart_title = chart.title or "Untitled"
            chart_type = type(chart).__name__
            # Try to get anchor position
            anchor = ""
            if hasattr(chart, "anchor") and chart.anchor:
                anchor = f" at:{chart.anchor}"
            charts_found.append(
                f"  {ws.title}: {chart_type} \"{chart_title}\"{anchor}"
            )

    if not charts_found:
        return "Charts: (none)"

    lines = [f"Charts ({len(charts_found)}):"]
    lines.extend(charts_found)
    return "\n".join(lines)


def _list_formulas(model: SheetsModel, index: SheetIndex) -> str:
    """List formula cells grouped by pattern."""
    wb = model.wb
    # pattern -> list of (sheet, addr, formula)
    patterns: dict[str, list[tuple[str, str, str]]] = {}

    for ws in wb.worksheets:
        sheet_name = ws.title
        bounds = index.get_bounds(sheet_name)
        if not bounds:
            continue
        min_row, min_col, max_row, max_col = bounds
        for row in ws.iter_rows(
            min_row=min_row, max_row=max_row,
            min_col=min_col, max_col=max_col,
        ):
            for cell in row:
                if isinstance(cell.value, str) and cell.value.startswith("="):
                    addr = format_cell_addr(cell.column, cell.row)
                    pattern = re.sub(r"\d+", "N", cell.value)
                    key = f"{sheet_name}|{pattern}"
                    if key not in patterns:
                        patterns[key] = []
                    patterns[key].append((sheet_name, addr, cell.value))

    if not patterns:
        return "Formulas: (none)"

    total = sum(len(v) for v in patterns.values())
    lines: list[str] = [f"Formulas ({total} cells, {len(patterns)} patterns):"]

    for key, entries in list(patterns.items())[:15]:
        sheet_name = entries[0][0]
        if len(entries) == 1:
            _, addr, formula = entries[0]
            lines.append(f"  {sheet_name}!{addr}  {formula}")
        else:
            first_addr = entries[0][1]
            last_addr = entries[-1][1]
            sample = entries[0][2]
            lines.append(
                f"  {sheet_name}!{first_addr}:{last_addr}  {sample}  ({len(entries)} cells)"
            )

    if len(patterns) > 15:
        lines.append(f"  ... +{len(patterns) - 15} more patterns")

    return "\n".join(lines)


def _list_styles(model: SheetsModel, index: SheetIndex) -> str:
    """List defined named styles."""
    wb = model.wb
    styles = list(wb.named_styles)

    if not styles:
        return "Named styles: (none)"

    lines: list[str] = [f"Named styles ({len(styles)}):"]
    for style in styles[:20]:
        name = style if isinstance(style, str) else style.name
        lines.append(f"  {name}")

    if len(styles) > 20:
        lines.append(f"  ... +{len(styles) - 20} more")

    return "\n".join(lines)


def _list_names(model: SheetsModel, index: SheetIndex) -> str:
    """List named ranges with scope."""
    wb = model.wb
    # wb.defined_names iterates over name strings; .values() gives DefinedName objects
    names = list(wb.defined_names.values()) if wb.defined_names else []

    if not names:
        return "Named ranges: (none)"

    lines: list[str] = [f"Named ranges ({len(names)}):"]
    for defn in names[:20]:
        scope = "workbook"
        if defn.localSheetId is not None:
            try:
                scope = wb.sheetnames[defn.localSheetId]
            except IndexError:
                scope = f"sheet#{defn.localSheetId}"

        # Get destinations
        dests = []
        try:
            for title, coord in defn.destinations:
                dests.append(f"{title}!{coord}")
        except Exception:
            dests.append(str(defn.value))

        dest_str = ", ".join(dests) if dests else str(defn.value)
        lines.append(f"  {defn.name}: {dest_str} (scope:{scope})")

    if len(names) > 20:
        lines.append(f"  ... +{len(names) - 20} more")

    return "\n".join(lines)


def _list_tables(model: SheetsModel, index: SheetIndex) -> str:
    """List Excel tables with range and style."""
    wb = model.wb
    tables_found: list[str] = []

    for ws in wb.worksheets:
        for table_name in ws.tables:
            table = ws.tables[table_name]
            style_name = ""
            if table.tableStyleInfo and table.tableStyleInfo.name:
                style_name = f" style:{table.tableStyleInfo.name}"
            tables_found.append(
                f"  {ws.title}: {table_name} {table.ref}{style_name}"
            )

    if not tables_found:
        return "Tables: (none)"

    lines = [f"Tables ({len(tables_found)}):"]
    lines.extend(tables_found)
    return "\n".join(lines)


# ---------------------------------------------------------------------------
# find — search cell values
# ---------------------------------------------------------------------------

_FIND_MAX_RESULTS = 50

def _query_find(args: str, model: SheetsModel, index: SheetIndex) -> str:
    """Search cell values or formulas across all sheets."""
    args = args.strip()
    if not args:
        return "! Usage: find TEXT | find formula:PATTERN"

    wb = model.wb

    # Determine search mode
    if args.lower().startswith("formula:"):
        pattern = args[8:]
        return _find_formulas(pattern, model, index)

    # Text search
    search_text = args.lower()
    results: list[str] = []

    for ws in wb.worksheets:
        sheet_name = ws.title
        bounds = index.get_bounds(sheet_name)
        if not bounds:
            continue
        min_row, min_col, max_row, max_col = bounds
        for row in ws.iter_rows(
            min_row=min_row, max_row=max_row,
            min_col=min_col, max_col=max_col,
        ):
            for cell in row:
                if cell.value is None:
                    continue
                val_str = str(cell.value)
                if search_text in val_str.lower():
                    addr = format_cell_addr(cell.column, cell.row)
                    display = _compact_value(cell.value)
                    results.append(f"  {sheet_name}!{addr} = {display}")
                    if len(results) >= _FIND_MAX_RESULTS:
                        break
            if len(results) >= _FIND_MAX_RESULTS:
                break
        if len(results) >= _FIND_MAX_RESULTS:
            break

    if not results:
        return f"find {args!r}: no matches"

    lines = [f"find {args!r}: {len(results)} match(es)"]
    lines.extend(results)
    if len(results) >= _FIND_MAX_RESULTS:
        lines.append(f"  ... capped at {_FIND_MAX_RESULTS} results")
    return "\n".join(lines)


def _find_formulas(pattern: str, model: SheetsModel, index: SheetIndex) -> str:
    """Search formula text across all sheets."""
    wb = model.wb
    search_pattern = pattern.lower()
    results: list[str] = []

    for ws in wb.worksheets:
        sheet_name = ws.title
        bounds = index.get_bounds(sheet_name)
        if not bounds:
            continue
        min_row, min_col, max_row, max_col = bounds
        for row in ws.iter_rows(
            min_row=min_row, max_row=max_row,
            min_col=min_col, max_col=max_col,
        ):
            for cell in row:
                if not isinstance(cell.value, str) or not cell.value.startswith("="):
                    continue
                if search_pattern in cell.value.lower():
                    addr = format_cell_addr(cell.column, cell.row)
                    results.append(f"  {sheet_name}!{addr} = {cell.value}")
                    if len(results) >= _FIND_MAX_RESULTS:
                        break
            if len(results) >= _FIND_MAX_RESULTS:
                break
        if len(results) >= _FIND_MAX_RESULTS:
            break

    if not results:
        return f"find formula:{pattern!r}: no matches"

    lines = [f"find formula:{pattern!r}: {len(results)} match(es)"]
    lines.extend(results)
    if len(results) >= _FIND_MAX_RESULTS:
        lines.append(f"  ... capped at {_FIND_MAX_RESULTS} results")
    return "\n".join(lines)


# ---------------------------------------------------------------------------
# Helpers
# ---------------------------------------------------------------------------

def _detect_formula_patterns(ws, bounds: tuple[int, int, int, int]) -> list[str]:
    """Detect formula patterns in a sheet for the plan query."""
    min_row, min_col, max_row, max_col = bounds
    patterns: dict[str, list[str]] = {}  # pattern -> [cell_addrs]

    for row in ws.iter_rows(
        min_row=min_row, max_row=max_row,
        min_col=min_col, max_col=max_col,
    ):
        for cell in row:
            if isinstance(cell.value, str) and cell.value.startswith("="):
                # Normalize formula pattern (strip row numbers for grouping)
                pattern = re.sub(r"\d+", "N", cell.value)
                addr = f"{index_to_col(cell.column)}{cell.row}"
                if pattern not in patterns:
                    patterns[pattern] = []
                patterns[pattern].append(addr)

    result = []
    for pattern, addrs in patterns.items():
        if len(addrs) == 1:
            result.append(f"{addrs[0]} {pattern.replace('N', 'N')}")
        else:
            # Show range
            first, last = addrs[0], addrs[-1]
            result.append(f"{first}:{last} pattern:{pattern}")

    return result[:5]  # Cap to avoid token explosion


def _compact_value(val) -> str:
    """Compact display of a cell value for peek/describe."""
    if val is None:
        return ""
    if isinstance(val, str):
        if val.startswith("="):
            return val
        return val[:30]
    if isinstance(val, bool):
        return str(val).upper()
    if isinstance(val, float):
        if val == int(val):
            return str(int(val))
        return f"{val:g}"
    if isinstance(val, datetime):
        return val.strftime("%Y-%m-%d")
    return str(val)


def _get_formula_groups(ws, bounds: tuple[int, int, int, int]) -> list[str]:
    """Get formula groupings for describe output."""
    min_row, min_col, max_row, max_col = bounds
    # pattern -> (first_addr, last_addr, count, sample)
    groups: dict[str, tuple[str, str, int, str]] = {}

    for row in ws.iter_rows(
        min_row=min_row, max_row=max_row,
        min_col=min_col, max_col=max_col,
    ):
        for cell in row:
            if isinstance(cell.value, str) and cell.value.startswith("="):
                pattern = re.sub(r"\d+", "N", cell.value)
                addr = format_cell_addr(cell.column, cell.row)
                if pattern not in groups:
                    groups[pattern] = (addr, addr, 1, cell.value)
                else:
                    first, _, count, sample = groups[pattern]
                    groups[pattern] = (first, addr, count + 1, sample)

    result: list[str] = []
    for pattern, (first, last, count, sample) in groups.items():
        if count == 1:
            result.append(f"{first}   {sample}")
        else:
            fill_note = "fill pattern" if count > 1 else ""
            result.append(f"{first}:{last}   {sample}  ({count} cells, {fill_note})")

    return result


def _describe_cf_rule(rule) -> str:
    """Describe a conditional formatting rule briefly."""
    rule_type = type(rule).__name__
    # Try to extract useful info
    if hasattr(rule, "type") and rule.type:
        return rule.type
    return rule_type



"""Editing operation handlers — copy, move, sort, insert/delete, remove."""

from __future__ import annotations

from copy import copy

from openpyxl.worksheet.worksheet import Worksheet

from fcp_core import OpResult, ParsedOp

from fcp_sheets.model.refs import (
    CellRef,
    RangeRef,
    col_to_index,
    index_to_col,
    parse_cell_ref,
    parse_range_ref,
)
from fcp_sheets.server.resolvers import SheetsOpContext, resolve_range_to_cells


def op_remove(op: ParsedOp, ctx: SheetsOpContext) -> OpResult:
    """Clear values in matched cells.

    Syntax: remove RANGE | remove @SELECTOR
    For MVP, accepts a range as positional[0].
    """
    range_str = None
    if op.positionals:
        range_str = op.positionals[0]
    elif op.selectors:
        # For MVP, treat @range:X as range X
        for sel in op.selectors:
            if sel.startswith("range:"):
                range_str = sel[6:]
                break
        if range_str is None:
            return OpResult(success=False, message="Unsupported selector for remove (MVP supports range only)")

    if not range_str:
        return OpResult(success=False, message="Usage: remove RANGE")

    ws = ctx.active_sheet
    count = 0

    # Try as range
    range_ref = parse_range_ref(range_str)
    if isinstance(range_ref, RangeRef):
        for row in range(range_ref.start.row, range_ref.end.row + 1):
            for col in range(range_ref.start.col, range_ref.end.col + 1):
                cell = ws.cell(row=row, column=col)
                if cell.value is not None:
                    cell.value = None
                    count += 1
        return OpResult(success=True, message=f"Cleared {count} cells in {range_str}", prefix="-")

    # Try as single cell
    cell_ref = parse_cell_ref(range_str)
    if cell_ref:
        cell = ws.cell(row=cell_ref.row, column=cell_ref.col)
        cell.value = None
        return OpResult(success=True, message=f"Cleared {range_str}", prefix="-")

    return OpResult(success=False, message=f"Invalid range: {range_str!r}")


def _copy_cell(src_ws: Worksheet, src_row: int, src_col: int,
               dst_ws: Worksheet, dst_row: int, dst_col: int) -> None:
    """Copy value and formatting from one cell to another."""
    src_cell = src_ws.cell(row=src_row, column=src_col)
    dst_cell = dst_ws.cell(row=dst_row, column=dst_col)

    # Copy value
    dst_cell.value = src_cell.value

    # Copy formatting attributes
    if src_cell.has_style:
        dst_cell.font = copy(src_cell.font)
        dst_cell.border = copy(src_cell.border)
        dst_cell.fill = copy(src_cell.fill)
        dst_cell.number_format = src_cell.number_format
        dst_cell.protection = copy(src_cell.protection)
        dst_cell.alignment = copy(src_cell.alignment)


def op_copy(op: ParsedOp, ctx: SheetsOpContext) -> OpResult:
    """Copy range to destination.

    Syntax: copy RANGE to:CELL [sheet:NAME]
    """
    if not op.positionals:
        return OpResult(success=False, message="Usage: copy RANGE to:CELL [sheet:NAME]")

    range_str = op.positionals[0]
    dest_str = op.params.get("to")
    if not dest_str:
        return OpResult(success=False, message="Missing 'to' parameter. Usage: copy RANGE to:CELL")

    # Parse source range
    range_ref = parse_range_ref(range_str)
    if not isinstance(range_ref, RangeRef):
        # Try single cell as 1x1 range
        cell_ref = parse_cell_ref(range_str)
        if cell_ref:
            range_ref = RangeRef(
                start=CellRef(col=cell_ref.col, row=cell_ref.row),
                end=CellRef(col=cell_ref.col, row=cell_ref.row),
            )
        else:
            return OpResult(success=False, message=f"Invalid source range: {range_str!r}")

    # Parse destination cell
    dest_ref = parse_cell_ref(dest_str)
    if not dest_ref:
        return OpResult(success=False, message=f"Invalid destination cell: {dest_str!r}")

    src_ws = ctx.active_sheet

    # Determine destination sheet
    sheet_name = op.params.get("sheet")
    if sheet_name:
        if sheet_name not in ctx.wb.sheetnames:
            return OpResult(success=False, message=f"Sheet '{sheet_name}' not found")
        dst_ws = ctx.wb[sheet_name]
    else:
        dst_ws = src_ws

    # Calculate offsets
    row_offset = dest_ref.row - range_ref.start.row
    col_offset = dest_ref.col - range_ref.start.col

    # Copy each cell
    count = 0
    for row in range(range_ref.start.row, range_ref.end.row + 1):
        for col in range(range_ref.start.col, range_ref.end.col + 1):
            _copy_cell(src_ws, row, col, dst_ws, row + row_offset, col + col_offset)
            count += 1

    # Update index for destination
    end_row = range_ref.end.row + row_offset
    end_col = range_ref.end.col + col_offset
    ctx.index.expand_bounds(dst_ws.title, dest_ref.row, dest_ref.col)
    ctx.index.expand_bounds(dst_ws.title, end_row, end_col)

    dest_label = f"{index_to_col(dest_ref.col)}{dest_ref.row}"
    sheet_info = f" on '{sheet_name}'" if sheet_name else ""
    return OpResult(success=True, message=f"Copied {range_str} to {dest_label}{sheet_info} ({count} cells)", prefix="+")


def op_move(op: ParsedOp, ctx: SheetsOpContext) -> OpResult:
    """Move range to destination (copy then clear source).

    Syntax: move RANGE to:CELL [sheet:NAME]
    """
    if not op.positionals:
        return OpResult(success=False, message="Usage: move RANGE to:CELL [sheet:NAME]")

    range_str = op.positionals[0]
    dest_str = op.params.get("to")
    if not dest_str:
        return OpResult(success=False, message="Missing 'to' parameter. Usage: move RANGE to:CELL")

    # Parse source range
    range_ref = parse_range_ref(range_str)
    if not isinstance(range_ref, RangeRef):
        cell_ref = parse_cell_ref(range_str)
        if cell_ref:
            range_ref = RangeRef(
                start=CellRef(col=cell_ref.col, row=cell_ref.row),
                end=CellRef(col=cell_ref.col, row=cell_ref.row),
            )
        else:
            return OpResult(success=False, message=f"Invalid source range: {range_str!r}")

    # Parse destination cell
    dest_ref = parse_cell_ref(dest_str)
    if not dest_ref:
        return OpResult(success=False, message=f"Invalid destination cell: {dest_str!r}")

    src_ws = ctx.active_sheet

    # Determine destination sheet
    sheet_name = op.params.get("sheet")
    if sheet_name:
        if sheet_name not in ctx.wb.sheetnames:
            return OpResult(success=False, message=f"Sheet '{sheet_name}' not found")
        dst_ws = ctx.wb[sheet_name]
    else:
        dst_ws = src_ws

    # Calculate offsets
    row_offset = dest_ref.row - range_ref.start.row
    col_offset = dest_ref.col - range_ref.start.col

    # Copy each cell first
    count = 0
    for row in range(range_ref.start.row, range_ref.end.row + 1):
        for col in range(range_ref.start.col, range_ref.end.col + 1):
            _copy_cell(src_ws, row, col, dst_ws, row + row_offset, col + col_offset)
            count += 1

    # Clear source cells (only if source != dest overlap area)
    for row in range(range_ref.start.row, range_ref.end.row + 1):
        for col in range(range_ref.start.col, range_ref.end.col + 1):
            dst_row = row + row_offset
            dst_col = col + col_offset
            # Only clear if this source cell doesn't overlap with a destination cell
            if (dst_ws is not src_ws or
                    dst_row < range_ref.start.row or dst_row > range_ref.end.row or
                    dst_col < range_ref.start.col or dst_col > range_ref.end.col or
                    row != dst_row or col != dst_col):
                src_cell = src_ws.cell(row=row, column=col)
                src_cell.value = None

    # Update index
    end_row = range_ref.end.row + row_offset
    end_col = range_ref.end.col + col_offset
    ctx.index.expand_bounds(dst_ws.title, dest_ref.row, dest_ref.col)
    ctx.index.expand_bounds(dst_ws.title, end_row, end_col)

    dest_label = f"{index_to_col(dest_ref.col)}{dest_ref.row}"
    sheet_info = f" on '{sheet_name}'" if sheet_name else ""
    return OpResult(success=True, message=f"Moved {range_str} to {dest_label}{sheet_info} ({count} cells)", prefix="*")


def op_sort(op: ParsedOp, ctx: SheetsOpContext) -> OpResult:
    """Sort range by column(s).

    Syntax: sort RANGE by:COL [dir:asc|desc] [by2:COL dir2:asc|desc]
    """
    if not op.positionals:
        return OpResult(success=False, message="Usage: sort RANGE by:COL [dir:asc|desc]")

    range_str = op.positionals[0]
    by_col = op.params.get("by")
    if not by_col:
        return OpResult(success=False, message="Missing 'by' parameter. Usage: sort RANGE by:COL")

    # Parse direction
    direction = op.params.get("dir", "asc").lower()
    if direction not in ("asc", "desc"):
        return OpResult(success=False, message=f"Invalid direction: {direction!r}. Use 'asc' or 'desc'")

    # Parse range
    range_ref = parse_range_ref(range_str)
    if not isinstance(range_ref, RangeRef):
        return OpResult(success=False, message=f"Invalid range: {range_str!r}")

    # Parse sort column — can be a letter like "B" or a full col ref
    sort_col_idx = col_to_index(by_col.upper()) if by_col.isalpha() else None
    if sort_col_idx is None:
        try:
            sort_col_idx = int(by_col)
        except ValueError:
            return OpResult(success=False, message=f"Invalid sort column: {by_col!r}")

    # Validate sort column is within range
    if sort_col_idx < range_ref.start.col or sort_col_idx > range_ref.end.col:
        return OpResult(
            success=False,
            message=f"Sort column {index_to_col(sort_col_idx)} is outside range {range_str}",
        )

    ws = ctx.active_sheet

    # Read all rows into memory
    rows_data: list[list] = []
    for row in range(range_ref.start.row, range_ref.end.row + 1):
        row_vals = []
        for col in range(range_ref.start.col, range_ref.end.col + 1):
            row_vals.append(ws.cell(row=row, column=col).value)
        rows_data.append(row_vals)

    # Sort key index (relative to range start col)
    key_idx = sort_col_idx - range_ref.start.col
    reverse = direction == "desc"

    # Secondary sort
    by2_col = op.params.get("by2")
    dir2 = op.params.get("dir2", "asc").lower()
    reverse2 = dir2 == "desc"

    if by2_col:
        sort_col2_idx = col_to_index(by2_col.upper()) if by2_col.isalpha() else None
        if sort_col2_idx is None:
            try:
                sort_col2_idx = int(by2_col)
            except ValueError:
                return OpResult(success=False, message=f"Invalid secondary sort column: {by2_col!r}")
        key2_idx = sort_col2_idx - range_ref.start.col

        def sort_key(row):
            v1 = row[key_idx]
            v2 = row[key2_idx]
            # Handle None — sort to end
            k1 = (0, v1) if v1 is not None else (1, "")
            k2 = (0, v2) if v2 is not None else (1, "")
            return k1, k2

        # For multi-key sort with different directions, use two-pass approach
        # Sort by secondary key first (stable sort), then primary
        if reverse != reverse2:
            rows_data.sort(key=lambda r: (0, r[key2_idx]) if r[key2_idx] is not None else (1, ""), reverse=reverse2)
            rows_data.sort(key=lambda r: (0, r[key_idx]) if r[key_idx] is not None else (1, ""), reverse=reverse)
        else:
            rows_data.sort(key=sort_key, reverse=reverse)
    else:
        def sort_key_single(row):
            v = row[key_idx]
            return (0, v) if v is not None else (1, "")

        rows_data.sort(key=sort_key_single, reverse=reverse)

    # Write sorted data back
    for i, row_vals in enumerate(rows_data):
        row = range_ref.start.row + i
        for j, val in enumerate(row_vals):
            col = range_ref.start.col + j
            ws.cell(row=row, column=col).value = val

    num_rows = range_ref.end.row - range_ref.start.row + 1
    sort_col_label = index_to_col(sort_col_idx)
    return OpResult(
        success=True,
        message=f"Sorted {range_str} by column {sort_col_label} ({direction}), {num_rows} rows",
        prefix="*",
    )


def op_insert_row(op: ParsedOp, ctx: SheetsOpContext) -> OpResult:
    """Insert rows.

    Syntax: insert-row ROW [count:N]
    """
    if not op.positionals:
        return OpResult(success=False, message="Usage: insert-row ROW [count:N]")

    try:
        row = int(op.positionals[0])
    except ValueError:
        return OpResult(success=False, message=f"Invalid row number: {op.positionals[0]!r}")

    if row < 1:
        return OpResult(success=False, message="Row number must be >= 1")

    count = 1
    if "count" in op.params:
        try:
            count = int(op.params["count"])
        except ValueError:
            return OpResult(success=False, message=f"Invalid count: {op.params['count']!r}")
        if count < 1:
            return OpResult(success=False, message="Count must be >= 1")

    ws = ctx.active_sheet
    ws.insert_rows(row, count)

    plural = "s" if count > 1 else ""
    return OpResult(success=True, message=f"Inserted {count} row{plural} at row {row}", prefix="+")


def op_insert_col(op: ParsedOp, ctx: SheetsOpContext) -> OpResult:
    """Insert columns.

    Syntax: insert-col COL [count:N]
    """
    if not op.positionals:
        return OpResult(success=False, message="Usage: insert-col COL [count:N]")

    col_str = op.positionals[0]

    # Accept column letter or number
    if col_str.isalpha():
        col_idx = col_to_index(col_str.upper())
    else:
        try:
            col_idx = int(col_str)
        except ValueError:
            return OpResult(success=False, message=f"Invalid column: {col_str!r}")

    if col_idx < 1:
        return OpResult(success=False, message="Column index must be >= 1")

    count = 1
    if "count" in op.params:
        try:
            count = int(op.params["count"])
        except ValueError:
            return OpResult(success=False, message=f"Invalid count: {op.params['count']!r}")
        if count < 1:
            return OpResult(success=False, message="Count must be >= 1")

    ws = ctx.active_sheet
    ws.insert_cols(col_idx, count)

    col_label = index_to_col(col_idx)
    plural = "s" if count > 1 else ""
    return OpResult(success=True, message=f"Inserted {count} column{plural} at column {col_label}", prefix="+")


def op_delete_row(op: ParsedOp, ctx: SheetsOpContext) -> OpResult:
    """Delete rows.

    Syntax: delete-row ROW [count:N]
    """
    if not op.positionals:
        return OpResult(success=False, message="Usage: delete-row ROW [count:N]")

    try:
        row = int(op.positionals[0])
    except ValueError:
        return OpResult(success=False, message=f"Invalid row number: {op.positionals[0]!r}")

    if row < 1:
        return OpResult(success=False, message="Row number must be >= 1")

    count = 1
    if "count" in op.params:
        try:
            count = int(op.params["count"])
        except ValueError:
            return OpResult(success=False, message=f"Invalid count: {op.params['count']!r}")
        if count < 1:
            return OpResult(success=False, message="Count must be >= 1")

    ws = ctx.active_sheet
    ws.delete_rows(row, count)

    plural = "s" if count > 1 else ""
    return OpResult(success=True, message=f"Deleted {count} row{plural} at row {row}", prefix="-")


def op_delete_col(op: ParsedOp, ctx: SheetsOpContext) -> OpResult:
    """Delete columns.

    Syntax: delete-col COL [count:N]
    """
    if not op.positionals:
        return OpResult(success=False, message="Usage: delete-col COL [count:N]")

    col_str = op.positionals[0]

    if col_str.isalpha():
        col_idx = col_to_index(col_str.upper())
    else:
        try:
            col_idx = int(col_str)
        except ValueError:
            return OpResult(success=False, message=f"Invalid column: {col_str!r}")

    if col_idx < 1:
        return OpResult(success=False, message="Column index must be >= 1")

    count = 1
    if "count" in op.params:
        try:
            count = int(op.params["count"])
        except ValueError:
            return OpResult(success=False, message=f"Invalid count: {op.params['count']!r}")
        if count < 1:
            return OpResult(success=False, message="Count must be >= 1")

    ws = ctx.active_sheet
    ws.delete_cols(col_idx, count)

    col_label = index_to_col(col_idx)
    plural = "s" if count > 1 else ""
    return OpResult(success=True, message=f"Deleted {count} column{plural} at column {col_label}", prefix="-")


HANDLERS: dict[str, callable] = {
    "remove": op_remove,
    "copy": op_copy,
    "move": op_move,
    "sort": op_sort,
    "insert-row": op_insert_row,
    "insert-col": op_insert_col,
    "delete-row": op_delete_row,
    "delete-col": op_delete_col,
}



"""Conditional formatting handlers — cond-fmt variants.

Supports: color-scale, data-bar, icon-set, cell-is, formula, duplicate, unique, top, bottom.
"""

from __future__ import annotations

from openpyxl.formatting.rule import (
    CellIsRule,
    ColorScaleRule,
    DataBarRule,
    FormulaRule,
    IconSetRule,
)
from openpyxl.styles import Font, PatternFill

from fcp_core import OpResult, ParsedOp

from fcp_sheets.lib.colors import parse_color
from fcp_sheets.server.resolvers import SheetsOpContext

# Operator mapping from DSL names to openpyxl operator names
_OP_MAP: dict[str, str] = {
    "gt": "greaterThan",
    "lt": "lessThan",
    "gte": "greaterThanOrEqual",
    "lte": "lessThanOrEqual",
    "eq": "equal",
    "neq": "notEqual",
    "ne": "notEqual",
    "between": "between",
    "not-between": "notBetween",
}

# Icon set style mapping
_ICON_MAP: dict[str, str] = {
    "arrows": "3Arrows",
    "flags": "3Flags",
    "traffic": "3TrafficLights1",
    "rating": "3Stars",
    "symbols": "3Symbols",
    "3arrows": "3Arrows",
    "3flags": "3Flags",
    "3traffic": "3TrafficLights1",
    "3stars": "3Stars",
    "3symbols": "3Symbols",
    "4arrows": "4Arrows",
    "4traffic": "4TrafficLights",
    "5arrows": "5Arrows",
    "5rating": "5Rating",
}


def _make_fill(color_hex: str | None) -> PatternFill | None:
    """Create a PatternFill from a hex color string, or None."""
    if color_hex is None:
        return None
    try:
        c = parse_color(color_hex)
        return PatternFill(start_color=c, end_color=c, fill_type="solid")
    except ValueError:
        return None


def _make_font(color_hex: str | None = None, bold: bool = False) -> Font | None:
    """Create a Font from optional color and bold flag, or None if both default."""
    if color_hex is None and not bold:
        return None
    font_color = None
    if color_hex:
        try:
            font_color = parse_color(color_hex)
        except ValueError:
            pass
    return Font(bold=bold, color=font_color)


def _cond_fmt_color_scale(op: ParsedOp, ctx: SheetsOpContext, range_str: str) -> OpResult:
    """cond-fmt RANGE color-scale [min-color:#HEX] [max-color:#HEX] [mid-color:#HEX]"""
    ws = ctx.active_sheet

    min_color = op.params.get("min-color", "#F8696B")
    max_color = op.params.get("max-color", "#63BE7B")
    mid_color = op.params.get("mid-color")

    try:
        min_c = parse_color(min_color)
        max_c = parse_color(max_color)
    except ValueError as e:
        return OpResult(success=False, message=str(e))

    if mid_color:
        try:
            mid_c = parse_color(mid_color)
        except ValueError as e:
            return OpResult(success=False, message=str(e))
        rule = ColorScaleRule(
            start_type="min", start_color=min_c,
            mid_type="percentile", mid_value=50, mid_color=mid_c,
            end_type="max", end_color=max_c,
        )
    else:
        rule = ColorScaleRule(
            start_type="min", start_color=min_c,
            end_type="max", end_color=max_c,
        )

    ws.conditional_formatting.add(range_str, rule)
    colors = "3-color" if mid_color else "2-color"
    return OpResult(success=True, message=f"Color scale ({colors}) on {range_str}", prefix="+")


def _cond_fmt_data_bar(op: ParsedOp, ctx: SheetsOpContext, range_str: str) -> OpResult:
    """cond-fmt RANGE data-bar [color:#HEX]"""
    ws = ctx.active_sheet

    color = op.params.get("color", "#638EC6")
    try:
        c = parse_color(color)
    except ValueError as e:
        return OpResult(success=False, message=str(e))

    rule = DataBarRule(start_type="min", end_type="max", color=c)
    ws.conditional_formatting.add(range_str, rule)

    return OpResult(success=True, message=f"Data bar on {range_str}", prefix="+")


def _cond_fmt_icon_set(op: ParsedOp, ctx: SheetsOpContext, range_str: str) -> OpResult:
    """cond-fmt RANGE icon-set [icons:arrows|flags|traffic|rating|symbols]"""
    ws = ctx.active_sheet

    icons_name = op.params.get("icons", "arrows").lower()
    icon_style = _ICON_MAP.get(icons_name)
    if icon_style is None:
        available = ", ".join(sorted(_ICON_MAP.keys()))
        return OpResult(success=False, message=f"Unknown icon set: {icons_name!r}. Available: {available}")

    rule = IconSetRule(icon_style=icon_style, type="percent", values=[0, 33, 67])
    ws.conditional_formatting.add(range_str, rule)

    return OpResult(success=True, message=f"Icon set ({icons_name}) on {range_str}", prefix="+")


def _cond_fmt_cell_is(op: ParsedOp, ctx: SheetsOpContext, range_str: str) -> OpResult:
    """cond-fmt RANGE cell-is OP VALUE [VALUE2] [fill:#HEX] [color:#HEX] [bold]"""
    ws = ctx.active_sheet

    # Expect: positionals = [RANGE, "cell-is", OP, VALUE, ...]
    if len(op.positionals) < 4:
        return OpResult(
            success=False,
            message="Usage: cond-fmt RANGE cell-is OP VALUE [VALUE2] [fill:#HEX] [color:#HEX] [bold]",
        )

    op_name = op.positionals[2].lower()
    mapped_op = _OP_MAP.get(op_name)
    if mapped_op is None:
        available = ", ".join(sorted(_OP_MAP.keys()))
        return OpResult(success=False, message=f"Unknown operator: {op_name!r}. Available: {available}")

    value1 = op.positionals[3]
    formula = [str(value1)]

    # For between/not-between, need a second value
    if mapped_op in ("between", "notBetween"):
        if len(op.positionals) < 5:
            return OpResult(success=False, message=f"Operator '{op_name}' requires two values")
        value2 = op.positionals[4]
        formula.append(str(value2))

    # Build formatting
    fill_color = op.params.get("fill")
    font_color = op.params.get("color")
    bold_flag = "bold" in [p.lower() for p in op.positionals]

    fill = _make_fill(fill_color)
    font = _make_font(font_color, bold_flag)

    kwargs: dict = {"operator": mapped_op, "formula": formula}
    if fill:
        kwargs["fill"] = fill
    if font:
        kwargs["font"] = font

    rule = CellIsRule(**kwargs)
    ws.conditional_formatting.add(range_str, rule)

    return OpResult(success=True, message=f"Cell-is ({op_name}) on {range_str}", prefix="+")


def _cond_fmt_formula(op: ParsedOp, ctx: SheetsOpContext, range_str: str) -> OpResult:
    """cond-fmt RANGE formula FORMULA [fill:#HEX] [color:#HEX] [bold]"""
    ws = ctx.active_sheet

    if len(op.positionals) < 3:
        return OpResult(
            success=False,
            message="Usage: cond-fmt RANGE formula FORMULA [fill:#HEX] [color:#HEX] [bold]",
        )

    formula_str = op.positionals[2]

    fill_color = op.params.get("fill")
    font_color = op.params.get("color")
    bold_flag = "bold" in [p.lower() for p in op.positionals]

    fill = _make_fill(fill_color)
    font = _make_font(font_color, bold_flag)

    kwargs: dict = {"formula": [formula_str]}
    if fill:
        kwargs["fill"] = fill
    if font:
        kwargs["font"] = font

    rule = FormulaRule(**kwargs)
    ws.conditional_formatting.add(range_str, rule)

    return OpResult(success=True, message=f"Formula rule on {range_str}", prefix="+")


def _cond_fmt_duplicate(op: ParsedOp, ctx: SheetsOpContext, range_str: str) -> OpResult:
    """cond-fmt RANGE duplicate [fill:#HEX]"""
    ws = ctx.active_sheet

    fill_color = op.params.get("fill", "#FFC7CE")
    fill = _make_fill(fill_color)

    # COUNTIF formula to detect duplicates
    # Use the first cell of the range as the reference
    first_cell = range_str.split(":")[0]
    formula_str = f"COUNTIF({range_str},{first_cell})>1"

    kwargs: dict = {"formula": [formula_str]}
    if fill:
        kwargs["fill"] = fill

    rule = FormulaRule(**kwargs)
    ws.conditional_formatting.add(range_str, rule)

    return OpResult(success=True, message=f"Duplicate highlighting on {range_str}", prefix="+")


def _cond_fmt_unique(op: ParsedOp, ctx: SheetsOpContext, range_str: str) -> OpResult:
    """cond-fmt RANGE unique [fill:#HEX]"""
    ws = ctx.active_sheet

    fill_color = op.params.get("fill", "#C6EFCE")
    fill = _make_fill(fill_color)

    first_cell = range_str.split(":")[0]
    formula_str = f"COUNTIF({range_str},{first_cell})=1"

    kwargs: dict = {"formula": [formula_str]}
    if fill:
        kwargs["fill"] = fill

    rule = FormulaRule(**kwargs)
    ws.conditional_formatting.add(range_str, rule)

    return OpResult(success=True, message=f"Unique highlighting on {range_str}", prefix="+")


def _cond_fmt_top(op: ParsedOp, ctx: SheetsOpContext, range_str: str) -> OpResult:
    """cond-fmt RANGE top N [fill:#HEX]"""
    ws = ctx.active_sheet

    if len(op.positionals) < 3:
        return OpResult(success=False, message="Usage: cond-fmt RANGE top N [fill:#HEX]")

    try:
        n = int(op.positionals[2])
    except ValueError:
        return OpResult(success=False, message=f"Invalid count: {op.positionals[2]!r}")

    fill_color = op.params.get("fill", "#C6EFCE")
    fill = _make_fill(fill_color)

    first_cell = range_str.split(":")[0]
    formula_str = f"RANK({first_cell},{range_str})<={n}"

    kwargs: dict = {"formula": [formula_str]}
    if fill:
        kwargs["fill"] = fill

    rule = FormulaRule(**kwargs)
    ws.conditional_formatting.add(range_str, rule)

    return OpResult(success=True, message=f"Top {n} on {range_str}", prefix="+")


def _cond_fmt_bottom(op: ParsedOp, ctx: SheetsOpContext, range_str: str) -> OpResult:
    """cond-fmt RANGE bottom N [fill:#HEX]"""
    ws = ctx.active_sheet

    if len(op.positionals) < 3:
        return OpResult(success=False, message="Usage: cond-fmt RANGE bottom N [fill:#HEX]")

    try:
        n = int(op.positionals[2])
    except ValueError:
        return OpResult(success=False, message=f"Invalid count: {op.positionals[2]!r}")

    fill_color = op.params.get("fill", "#FFC7CE")
    fill = _make_fill(fill_color)

    first_cell = range_str.split(":")[0]
    formula_str = f"RANK({first_cell},{range_str})>=COUNT({range_str})-{n}+1"

    kwargs: dict = {"formula": [formula_str]}
    if fill:
        kwargs["fill"] = fill

    rule = FormulaRule(**kwargs)
    ws.conditional_formatting.add(range_str, rule)

    return OpResult(success=True, message=f"Bottom {n} on {range_str}", prefix="+")


# Sub-type dispatch
_COND_FMT_TYPES: dict[str, callable] = {
    "color-scale": _cond_fmt_color_scale,
    "data-bar": _cond_fmt_data_bar,
    "icon-set": _cond_fmt_icon_set,
    "cell-is": _cond_fmt_cell_is,
    "formula": _cond_fmt_formula,
    "duplicate": _cond_fmt_duplicate,
    "unique": _cond_fmt_unique,
    "top": _cond_fmt_top,
    "bottom": _cond_fmt_bottom,
}


def op_cond_fmt(op: ParsedOp, ctx: SheetsOpContext) -> OpResult:
    """Main cond-fmt verb dispatcher.

    Syntax: cond-fmt RANGE TYPE [params...]
    """
    if len(op.positionals) < 2:
        return OpResult(
            success=False,
            message="Usage: cond-fmt RANGE TYPE [params...]",
        )

    range_str = op.positionals[0]
    fmt_type = op.positionals[1].lower()

    handler = _COND_FMT_TYPES.get(fmt_type)
    if handler is None:
        available = ", ".join(sorted(_COND_FMT_TYPES.keys()))
        return OpResult(success=False, message=f"Unknown cond-fmt type: {fmt_type!r}. Available: {available}")

    return handler(op, ctx, range_str)


HANDLERS: dict[str, callable] = {
    "cond-fmt": op_cond_fmt,
}



"""Data validation handlers — validate variants.

Supports: list, number, date, length, custom, off.
"""

from __future__ import annotations

from openpyxl.worksheet.datavalidation import DataValidation

from fcp_core import OpResult, ParsedOp

from fcp_sheets.server.resolvers import SheetsOpContext

# Operator mapping
_OP_MAP: dict[str, str] = {
    "gt": "greaterThan",
    "lt": "lessThan",
    "gte": "greaterThanOrEqual",
    "lte": "lessThanOrEqual",
    "eq": "equal",
    "neq": "notEqual",
    "ne": "notEqual",
    "between": "between",
    "not-between": "notBetween",
}

# Type mapping from DSL names to openpyxl validation types
_TYPE_MAP: dict[str, str] = {
    "number": "decimal",
    "whole": "whole",
    "decimal": "decimal",
    "date": "date",
    "length": "textLength",
    "text-length": "textLength",
    "time": "time",
}


def _validate_list(op: ParsedOp, ctx: SheetsOpContext, range_str: str) -> OpResult:
    """validate RANGE list VALUES|range:RANGE

    Values can be comma-separated inline or a range reference.
    """
    ws = ctx.active_sheet

    # Check for range: param (list from range)
    list_range = op.params.get("range")
    if list_range:
        formula1 = list_range
    else:
        # Inline values from remaining positionals
        if len(op.positionals) < 3:
            return OpResult(
                success=False,
                message="Usage: validate RANGE list Val1,Val2,Val3 | validate RANGE list range:A1:A10",
            )
        # Reconstruct the value string — the tokenizer may have split
        # multi-word items on spaces (e.g. "Exceeded,On Track,At Risk"
        # becomes positionals ["Exceeded,On", "Track,At", "Risk"]).
        raw = " ".join(op.positionals[2:])
        if "," in raw:
            # Comma-delimited: split on commas, strip whitespace per item
            items = [item.strip() for item in raw.split(",") if item.strip()]
            values = ",".join(items)
        else:
            # Space-delimited: each positional is a separate item
            values = ",".join(op.positionals[2:])
        formula1 = f'"{values}"'

    dv = DataValidation(type="list", formula1=formula1, allow_blank=True)
    dv.add(range_str)
    ws.add_data_validation(dv)

    return OpResult(success=True, message=f"List validation on {range_str}", prefix="+")


def _validate_typed(op: ParsedOp, ctx: SheetsOpContext, range_str: str) -> OpResult:
    """validate RANGE number|date|length OP VALUE [VALUE2]"""
    ws = ctx.active_sheet

    if len(op.positionals) < 4:
        return OpResult(
            success=False,
            message="Usage: validate RANGE number|date|length OP VALUE [VALUE2]",
        )

    type_name = op.positionals[1].lower()
    val_type = _TYPE_MAP.get(type_name)
    if val_type is None:
        available = ", ".join(sorted(_TYPE_MAP.keys()))
        return OpResult(success=False, message=f"Unknown validation type: {type_name!r}. Available: {available}")

    op_name = op.positionals[2].lower()
    mapped_op = _OP_MAP.get(op_name)
    if mapped_op is None:
        available = ", ".join(sorted(_OP_MAP.keys()))
        return OpResult(success=False, message=f"Unknown operator: {op_name!r}. Available: {available}")

    value1 = op.positionals[3]

    kwargs: dict = {
        "type": val_type,
        "operator": mapped_op,
        "formula1": str(value1),
        "allow_blank": True,
    }

    # For between/not-between, need a second value
    if mapped_op in ("between", "notBetween"):
        if len(op.positionals) < 5:
            return OpResult(success=False, message=f"Operator '{op_name}' requires two values")
        value2 = op.positionals[4]
        kwargs["formula2"] = str(value2)

    dv = DataValidation(**kwargs)
    dv.add(range_str)
    ws.add_data_validation(dv)

    return OpResult(success=True, message=f"{type_name.capitalize()} validation ({op_name}) on {range_str}", prefix="+")


def _validate_custom(op: ParsedOp, ctx: SheetsOpContext, range_str: str) -> OpResult:
    """validate RANGE custom FORMULA"""
    ws = ctx.active_sheet

    if len(op.positionals) < 3:
        return OpResult(success=False, message="Usage: validate RANGE custom FORMULA")

    formula = op.positionals[2]

    dv = DataValidation(type="custom", formula1=formula, allow_blank=True)
    dv.add(range_str)
    ws.add_data_validation(dv)

    return OpResult(success=True, message=f"Custom validation on {range_str}", prefix="+")


def _validate_off(op: ParsedOp, ctx: SheetsOpContext) -> OpResult:
    """validate off RANGE"""
    if len(op.positionals) < 2:
        return OpResult(success=False, message="Usage: validate off RANGE")

    range_str = op.positionals[1]
    ws = ctx.active_sheet

    # Remove data validations that cover this range
    to_remove = []
    for dv in ws.data_validations.dataValidation:
        # Check if any of the dv's cell ranges overlap with the target range
        for cr in dv.cells.ranges:
            if str(cr) == range_str:
                to_remove.append(dv)
                break

    if not to_remove:
        return OpResult(success=False, message=f"No validation found on {range_str}")

    for dv in to_remove:
        ws.data_validations.dataValidation.remove(dv)

    count = len(to_remove)
    return OpResult(
        success=True,
        message=f"Removed {count} validation(s) from {range_str}",
        prefix="-",
    )


# Sub-type dispatch
_VALIDATE_TYPES: dict[str, callable] = {
    "list": _validate_list,
    "custom": _validate_custom,
    # Typed validations handled by _validate_typed
}


def op_validate(op: ParsedOp, ctx: SheetsOpContext) -> OpResult:
    """Main validate verb dispatcher.

    Syntax: validate RANGE TYPE [params...] | validate off RANGE
    """
    if not op.positionals:
        return OpResult(success=False, message="Usage: validate RANGE TYPE [params...] | validate off RANGE")

    # Handle "validate off RANGE"
    if op.positionals[0].lower() == "off":
        return _validate_off(op, ctx)

    if len(op.positionals) < 2:
        return OpResult(success=False, message="Usage: validate RANGE TYPE [params...]")

    range_str = op.positionals[0]
    val_type = op.positionals[1].lower()

    # Check specific handlers first
    handler = _VALIDATE_TYPES.get(val_type)
    if handler:
        return handler(op, ctx, range_str)

    # Check if it's a typed validation (number, date, length, etc.)
    if val_type in _TYPE_MAP:
        return _validate_typed(op, ctx, range_str)

    available = ", ".join(sorted(list(_VALIDATE_TYPES.keys()) + list(_TYPE_MAP.keys())))
    return OpResult(success=False, message=f"Unknown validation type: {val_type!r}. Available: {available}")


HANDLERS: dict[str, callable] = {
    "validate": op_validate,
}



#!/usr/bin/env python3
"""Showdown v2 Auditor — checks both FCP and Raw xlsx files against the spec."""

import json
import sys
from pathlib import Path
from openpyxl import load_workbook
from openpyxl.utils import get_column_letter

DIR = Path(__file__).parent


def audit_file(path: str, label: str) -> dict:
    """Audit a single xlsx file. Returns {checks: [...], score: (pass, total)}."""
    wb = load_workbook(path, data_only=False)
    checks = []

    def check(category, name, passed, detail=""):
        checks.append({
            "category": category,
            "name": name,
            "passed": passed,
            "detail": detail
        })

    sheet_names = wb.sheetnames

    # ═══════════════════════════════════════════════════════
    # PHASE 1 CHECKS
    # ═══════════════════════════════════════════════════════

    # --- Structure ---
    expected_sheets_p1 = ["Executive Summary", "Fund Performance", "Holdings", "Portfolio Detail", "Cash Flows", "Scenario Analysis"]
    # After Phase 2, "Holdings" is renamed to "Portfolio Detail" and "Sector Allocation" is added
    has_portfolio_detail = "Portfolio Detail" in sheet_names
    has_holdings = "Holdings" in sheet_names
    holdings_name = "Portfolio Detail" if has_portfolio_detail else "Holdings"

    expected_final = ["Executive Summary", "Fund Performance", holdings_name, "Cash Flows", "Scenario Analysis", "Sector Allocation"]

    check("Structure", "Has Executive Summary sheet", "Executive Summary" in sheet_names)
    check("Structure", "Has Fund Performance sheet", "Fund Performance" in sheet_names)
    check("Structure", "Has Holdings/Portfolio Detail sheet", has_holdings or has_portfolio_detail,
          f"Found: {holdings_name}")
    check("Structure", "Has Cash Flows sheet", "Cash Flows" in sheet_names)
    check("Structure", "Has Scenario Analysis sheet", "Scenario Analysis" in sheet_names)
    check("Structure", "No default 'Sheet' tab", "Sheet" not in sheet_names,
          f"Sheets: {sheet_names}")

    # --- Executive Summary ---
    if "Executive Summary" in sheet_names:
        ws = wb["Executive Summary"]

        # Title merge
        merged = [str(m) for m in ws.merged_cells.ranges]
        check("ExecSummary", "Title merge A1:F1", any("A1:F1" in m for m in merged),
              f"Merges: {merged[:5]}")

        # Fund Overview data
        check("ExecSummary", "Fund Size = 500000000", ws["B5"].value == 500000000,
              f"B5={ws['B5'].value}")
        check("ExecSummary", "Vintage Year = 2021", ws["B6"].value == 2021,
              f"B6={ws['B6'].value}")
        check("ExecSummary", "Investment Period text", ws["B7"].value == "2021–2024" or ws["B7"].value == "2021-2024",
              f"B7={ws['B7'].value}")
        check("ExecSummary", "Invested Capital = 385000000", ws["B8"].value == 385000000,
              f"B8={ws['B8'].value}")
        check("ExecSummary", "Realized Value = 142000000", ws["B9"].value == 142000000,
              f"B9={ws['B9'].value}")
        check("ExecSummary", "Unrealized Value = 498000000", ws["B10"].value == 498000000,
              f"B10={ws['B10'].value}")
        # Total Value formula
        b11 = ws["B11"].value
        check("ExecSummary", "Total Value formula (=B9+B10)",
              isinstance(b11, str) and "B9" in b11 and "B10" in b11,
              f"B11={b11}")

        # Performance metrics
        check("ExecSummary", "Net IRR = 0.218", ws["B14"].value == 0.218,
              f"B14={ws['B14'].value}")
        check("ExecSummary", "Gross IRR = 0.267", ws["B15"].value == 0.267,
              f"B15={ws['B15'].value}")
        check("ExecSummary", "Net MOIC = 1.66", ws["B16"].value == 1.66,
              f"B16={ws['B16'].value}")
        # DPI formula
        b17 = ws["B17"].value
        check("ExecSummary", "DPI formula (=B9/B8)",
              isinstance(b17, str) and "B9" in str(b17) and "B8" in str(b17),
              f"B17={b17}")

        # Portfolio Snapshot
        check("ExecSummary", "Active Companies = 8", ws["B22"].value == 8,
              f"B22={ws['B22'].value}")
        check("ExecSummary", "Fully Exited = 2", ws["B23"].value == 2,
              f"B23={ws['B23'].value}")

        # Named ranges
        defined_names = {dn.name: str(dn.value) for dn in wb.defined_names.values()}
        check("ExecSummary", "Named range FundSize", "FundSize" in defined_names,
              f"Names: {list(defined_names.keys())[:10]}")
        check("ExecSummary", "Named range InvestedCapital", "InvestedCapital" in defined_names)
        check("ExecSummary", "Named range TotalValue", "TotalValue" in defined_names)

        # Protection
        check("ExecSummary", "Sheet protected", ws.protection.sheet,
              f"protection.sheet={ws.protection.sheet}")

    # --- Fund Performance ---
    if "Fund Performance" in sheet_names:
        ws = wb["Fund Performance"]

        # Title merge
        merged = [str(m) for m in ws.merged_cells.ranges]
        check("FundPerf", "Title merge A1:H1", any("A1:H1" in m or "A1:I1" in m for m in merged),
              f"Merges: {merged[:5]}")

        # Data values (row 3 = 2021)
        check("FundPerf", "2021 Contributions = -125000000", ws["B3"].value == -125000000,
              f"B3={ws['B3'].value}")
        check("FundPerf", "2022 Distributions = 12000000", ws["C4"].value == 12000000,
              f"C4={ws['C4'].value}")
        check("FundPerf", "2025 Net IRR = 0.218", ws["F7"].value == 0.218,
              f"F7={ws['F7'].value}")
        check("FundPerf", "2025 MOIC = 1.66", ws["G7"].value == 1.66,
              f"G7={ws['G7'].value}")

        # Net Cash Flow formulas
        d3 = ws["D3"].value
        check("FundPerf", "Net Cash Flow formula D3 (=B3+C3)",
              isinstance(d3, str) and "B3" in str(d3) and "C3" in str(d3),
              f"D3={d3}")

        # Total row formulas
        b8 = ws["B8"].value
        check("FundPerf", "Total Contributions formula (SUM)",
              isinstance(b8, str) and "SUM" in str(b8).upper(),
              f"B8={b8}")

        # Freeze pane
        check("FundPerf", "Freeze pane at A3", ws.freeze_panes == "A3",
              f"freeze={ws.freeze_panes}")

        # Charts
        chart_count = len(ws._charts)
        check("FundPerf", "Has 2 charts", chart_count >= 2,
              f"Charts: {chart_count}")

    # --- Holdings / Portfolio Detail ---
    if holdings_name in sheet_names:
        ws = wb[holdings_name]

        # Company data (row 3 = Helios Energy)
        check("Holdings", "Helios Energy in A3",
              ws["A3"].value and "Helios" in str(ws["A3"].value),
              f"A3={ws['A3'].value}")
        check("Holdings", "Helios Cost Basis = 65000000", ws["D3"].value == 65000000,
              f"D3={ws['D3'].value}")
        check("Holdings", "Helios Fair Value = 142000000", ws["E3"].value == 142000000,
              f"E3={ws['E3'].value}")

        # MOIC formula
        f3 = ws["F3"].value
        check("Holdings", "MOIC formula F3 (=E3/D3)",
              isinstance(f3, str) and "E3" in str(f3) and "D3" in str(f3),
              f"F3={f3}")

        # Written Off company
        check("Holdings", "Bolt Payments status = Written Off",
              ws["K11"].value == "Written Off",
              f"K11={ws['K11'].value}")

        # Apex Dynamics (Exited)
        check("Holdings", "Apex Dynamics status = Exited",
              ws["K12"].value == "Exited",
              f"K12={ws['K12'].value}")

        # Totals row formulas
        d13 = ws["D13"].value
        check("Holdings", "Totals Cost Basis formula (SUM)",
              isinstance(d13, str) and "SUM" in str(d13).upper(),
              f"D13={d13}")

        # Freeze pane
        check("Holdings", "Freeze pane at A3", ws.freeze_panes == "A3",
              f"freeze={ws.freeze_panes}")

        # Filter
        check("Holdings", "AutoFilter enabled", ws.auto_filter.ref is not None,
              f"filter={ws.auto_filter.ref}")

        # Chart
        chart_count = len(ws._charts)
        check("Holdings", "Has chart (bubble/scatter)", chart_count >= 1,
              f"Charts: {chart_count}")

    # --- Cash Flows ---
    if "Cash Flows" in sheet_names:
        ws = wb["Cash Flows"]

        # Title merge
        merged = [str(m) for m in ws.merged_cells.ranges]
        check("CashFlows", "Title merge A1:F1 or A1:G1",
              any("A1:F1" in m or "A1:G1" in m for m in merged),
              f"Merges: {merged[:5]}")

        # Data values
        check("CashFlows", "Q1 2021 Contributions = -50000000", ws["B3"].value == -50000000,
              f"B3={ws['B3'].value}")
        check("CashFlows", "Q1 2021 NAV = 48000000", ws["F3"].value == 48000000,
              f"F3={ws['F3'].value}")
        check("CashFlows", "Q4 2025 NAV = 640000000", ws["F22"].value == 640000000,
              f"F22={ws['F22'].value}")

        # Net Cash Flow formula
        d3 = ws["D3"].value
        check("CashFlows", "Net Cash Flow formula D3 (=B3+C3)",
              isinstance(d3, str) and "B3" in str(d3) and "C3" in str(d3),
              f"D3={d3}")

        # Cumulative Cash Flow formula
        e4 = ws["E4"].value
        check("CashFlows", "Cumulative formula E4 (=E3+D4)",
              isinstance(e4, str) and "E3" in str(e4) and "D4" in str(e4),
              f"E4={e4}")

        # Total row
        b23 = ws["B23"].value
        check("CashFlows", "Total Contributions formula (SUM)",
              isinstance(b23, str) and "SUM" in str(b23).upper(),
              f"B23={b23}")

        # Named range LatestNAV
        check("CashFlows", "Named range LatestNAV", "LatestNAV" in defined_names,
              f"Names: {list(defined_names.keys())[:10]}")

        # Freeze
        check("CashFlows", "Freeze pane at A3", ws.freeze_panes == "A3",
              f"freeze={ws.freeze_panes}")

        # Charts
        chart_count = len(ws._charts)
        check("CashFlows", "Has 2 charts", chart_count >= 2,
              f"Charts: {chart_count}")

    # --- Scenario Analysis ---
    if "Scenario Analysis" in sheet_names:
        ws = wb["Scenario Analysis"]

        # Assumptions data
        check("ScenarioAnalysis", "Revenue Growth Base = 0.25", ws["B5"].value == 0.25,
              f"B5={ws['B5'].value}")
        check("ScenarioAnalysis", "Revenue Growth Bull = 0.40", ws["C5"].value == 0.40 or ws["C5"].value == 0.4,
              f"C5={ws['C5'].value}")
        check("ScenarioAnalysis", "Loss Rate Bear = 0.15", ws["D9"].value == 0.15,
              f"D9={ws['D9'].value}")

        # Cross-sheet formula (Portfolio Revenue)
        b13 = ws["B13"].value
        check("ScenarioAnalysis", "Portfolio Revenue cross-sheet formula",
              isinstance(b13, str) and ("Holdings" in str(b13) or "Portfolio Detail" in str(b13)),
              f"B13={b13}")

        # Implied MOIC cross-sheet formula
        b18 = ws["B18"].value
        check("ScenarioAnalysis", "Implied MOIC cross-sheet formula",
              isinstance(b18, str) and "Executive Summary" in str(b18),
              f"B18={b18}")

        # Chart
        chart_count = len(ws._charts)
        check("ScenarioAnalysis", "Has MOIC comparison chart", chart_count >= 1,
              f"Charts: {chart_count}")

    # ═══════════════════════════════════════════════════════
    # PHASE 2 CHECKS
    # ═══════════════════════════════════════════════════════

    # --- Cash Flows Total Return column ---
    if "Cash Flows" in sheet_names:
        ws = wb["Cash Flows"]
        g2 = ws["G2"].value
        check("Phase2-CF", "Total Return header in G2",
              g2 and "Total Return" in str(g2),
              f"G2={g2}")

        g3 = ws["G3"].value
        check("Phase2-CF", "Total Return formula G3 (Cumulative + NAV)",
              isinstance(g3, str) and ("E3" in str(g3) or "F3" in str(g3)),
              f"G3={g3}")

        g22 = ws["G22"].value
        check("Phase2-CF", "Total Return has value in G22",
              g22 is not None,
              f"G22={g22}")

        # Total row: should be =G22, not SUM
        g23 = ws["G23"].value
        check("Phase2-CF", "Total Return total = G22 ref (not SUM)",
              isinstance(g23, str) and "G22" in str(g23),
              f"G23={g23}")

    # --- Holdings renamed to Portfolio Detail ---
    check("Phase2-Holdings", "Sheet renamed to 'Portfolio Detail'", has_portfolio_detail,
          f"Found: {'Portfolio Detail' if has_portfolio_detail else 'Holdings'}")

    # --- Investment Thesis column (L) ---
    if holdings_name in sheet_names:
        ws = wb[holdings_name]
        l2 = ws["L2"].value
        check("Phase2-Holdings", "Investment Thesis header in L2",
              l2 and "Thesis" in str(l2),
              f"L2={l2}")

        l3 = ws["L3"].value
        check("Phase2-Holdings", "Helios thesis = 'Grid-scale battery storage leader'",
              l3 and "battery" in str(l3).lower(),
              f"L3={l3}")

        l6 = ws["L6"].value
        check("Phase2-Holdings", "Prism thesis = 'AI-native business intelligence'",
              l6 and "AI" in str(l6),
              f"L6={l6}")

    # --- Watchlist column (M) ---
    if holdings_name in sheet_names:
        ws = wb[holdings_name]
        m2 = ws["M2"].value
        check("Phase2-Holdings", "Watchlist header in M2",
              m2 and "Watchlist" in str(m2),
              f"M2={m2}")

        m3 = ws["M3"].value
        check("Phase2-Holdings", "Helios Watchlist = Green",
              m3 and str(m3).strip().lower() == "green",
              f"M3={m3}")

        m4 = ws["M4"].value
        check("Phase2-Holdings", "NovaBio Watchlist = Red",
              m4 and str(m4).strip().lower() == "red",
              f"M4={m4}")

        m7 = ws["M7"].value
        check("Phase2-Holdings", "Stratos Watchlist = Yellow",
              m7 and str(m7).strip().lower() == "yellow",
              f"M7={m7}")

    # --- Sector Allocation sheet ---
    check("Phase2-Sector", "Has Sector Allocation sheet", "Sector Allocation" in sheet_names,
          f"Sheets: {sheet_names}")

    if "Sector Allocation" in sheet_names:
        ws = wb["Sector Allocation"]

        # Title
        a1 = ws["A1"].value
        check("Phase2-Sector", "Title contains 'Sector Allocation'",
              a1 and "Sector" in str(a1),
              f"A1={a1}")

        # Header row
        a2 = ws["A2"].value
        check("Phase2-Sector", "Header row exists (A2=Sector)",
              a2 and "Sector" in str(a2),
              f"A2={a2}")

        # Software sector (should be first or present)
        found_software = False
        for row in range(3, 12):
            val = ws[f"A{row}"].value
            if val and "Software" in str(val):
                found_software = True
                # Check company count
                b_val = ws[f"B{row}"].value
                check("Phase2-Sector", "Software companies count = 2",
                      b_val == 2,
                      f"B{row}={b_val}")
                break
        check("Phase2-Sector", "Software sector present", found_software)

        # Chart
        chart_count = len(ws._charts)
        check("Phase2-Sector", "Has doughnut/pie chart", chart_count >= 1,
              f"Charts: {chart_count}")

    # ═══════════════════════════════════════════════════════
    # PHASE 3 CHECKS
    # ═══════════════════════════════════════════════════════

    # --- Navigation hyperlinks ---
    if "Executive Summary" in sheet_names:
        ws = wb["Executive Summary"]
        a26 = ws["A26"].value
        check("Phase3", "Navigation section header A26",
              a26 and "Navigation" in str(a26),
              f"A26={a26}")

        # Check for hyperlinks in A27-A31
        has_links = False
        for row in range(27, 32):
            cell = ws[f"A{row}"]
            if cell.hyperlink is not None:
                has_links = True
                break
        check("Phase3", "Has navigation hyperlinks (A27-A31)", has_links)

    # --- Data bars on Fund Performance distributions ---
    if "Fund Performance" in sheet_names:
        ws = wb["Fund Performance"]
        cf_rules = ws.conditional_formatting._cf_rules if hasattr(ws.conditional_formatting, '_cf_rules') else list(ws.conditional_formatting)
        has_data_bars = False
        for rule in cf_rules:
            rule_str = str(rule)
            if "dataBar" in rule_str.lower() or "DataBar" in str(type(rule)):
                has_data_bars = True
                break
            # Check rule object
            if hasattr(rule, 'rule'):
                if hasattr(rule.rule, 'dataBar') and rule.rule.dataBar is not None:
                    has_data_bars = True
                    break
        # Also check via the conditional formatting list
        for cf in ws.conditional_formatting:
            for rule in cf.rules:
                if rule.dataBar is not None:
                    has_data_bars = True
                    break
        check("Phase3", "Data bars on Fund Performance distributions", has_data_bars)

    # --- Page footers ---
    footer_count = 0
    for sn in sheet_names:
        ws = wb[sn]
        footer = ws.oddFooter
        if footer and ("Meridian" in str(footer) or "Confidential" in str(footer)):
            footer_count += 1
    check("Phase3", "Page footers set on sheets",
          footer_count >= 3,
          f"Sheets with footers: {footer_count}/{len(sheet_names)}")

    # --- Print title rows ---
    if holdings_name in sheet_names:
        ws = wb[holdings_name]
        check("Phase3", f"Print titles on {holdings_name}",
              ws.print_title_rows is not None,
              f"print_title_rows={ws.print_title_rows}")

    if "Cash Flows" in sheet_names:
        ws = wb["Cash Flows"]
        check("Phase3", "Print titles on Cash Flows",
              ws.print_title_rows is not None,
              f"print_title_rows={ws.print_title_rows}")

    # --- LatestMOIC named range ---
    check("Phase3", "Named range LatestMOIC", "LatestMOIC" in defined_names,
          f"Names: {list(defined_names.keys())}")

    # ═══════════════════════════════════════════════════════
    # Summary
    # ═══════════════════════════════════════════════════════
    passed = sum(1 for c in checks if c["passed"])
    total = len(checks)

    return {
        "label": label,
        "checks": checks,
        "passed": passed,
        "total": total,
    }


def print_results(result):
    """Print audit results."""
    print(f"\n{'=' * 65}")
    print(f"AUDIT: {result['label']}")
    print(f"{'=' * 65}")

    current_cat = None
    for c in result["checks"]:
        if c["category"] != current_cat:
            current_cat = c["category"]
            print(f"\n  [{current_cat}]")

        icon = "PASS" if c["passed"] else "FAIL"
        detail = f" — {c['detail']}" if c["detail"] and not c["passed"] else ""
        print(f"    [{icon}] {c['name']}{detail}")

    print(f"\n  Score: {result['passed']}/{result['total']} ({100*result['passed']/result['total']:.0f}%)")


def main():
    fcp_path = DIR / "v2-fcp-result.xlsx"
    raw_path = DIR / "v2-raw-result.xlsx"

    results = {}

    if fcp_path.exists():
        results["fcp"] = audit_file(str(fcp_path), "FCP (fcp-sheets DSL)")
        print_results(results["fcp"])
    else:
        print(f"SKIP: {fcp_path} not found")

    if raw_path.exists():
        results["raw"] = audit_file(str(raw_path), "Raw (openpyxl Python)")
        print_results(results["raw"])
    else:
        print(f"SKIP: {raw_path} not found")

    # Comparison
    if len(results) == 2:
        print(f"\n{'=' * 65}")
        print("COMPARISON SUMMARY")
        print(f"{'=' * 65}")
        print(f"  {'Check':<45} {'FCP':>5} {'Raw':>5}")
        print(f"  {'─' * 55}")

        fcp_checks = {c["name"]: c["passed"] for c in results["fcp"]["checks"]}
        raw_checks = {c["name"]: c["passed"] for c in results["raw"]["checks"]}

        all_names = list(dict.fromkeys(
            [c["name"] for c in results["fcp"]["checks"]] +
            [c["name"] for c in results["raw"]["checks"]]
        ))

        disagree = 0
        for name in all_names:
            fcp_p = fcp_checks.get(name, False)
            raw_p = raw_checks.get(name, False)
            fcp_icon = "PASS" if fcp_p else "FAIL"
            raw_icon = "PASS" if raw_p else "FAIL"
            marker = " <<" if fcp_p != raw_p else ""
            print(f"  {name:<45} {fcp_icon:>5} {raw_icon:>5}{marker}")
            if fcp_p != raw_p:
                disagree += 1

        fcp_r = results["fcp"]
        raw_r = results["raw"]
        print(f"\n  TOTAL: FCP {fcp_r['passed']}/{fcp_r['total']} vs Raw {raw_r['passed']}/{raw_r['total']}")
        if disagree:
            print(f"  Disagreements: {disagree} checks (marked <<)")

    # Save JSON
    output_path = DIR / "v2-audit-results.json"
    with open(output_path, "w") as f:
        json.dump(results, f, indent=2, default=str)
    print(f"\nResults saved to {output_path}")


if __name__ == "__main__":
    main()
